import time

import numpy as np
from bidict import FrozenOrderedBidict

from kamzik3 import DeviceError
from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.snippets.snippetsDecorators import expose_method

try:
    from PyDAQmx import *
except ImportError:
    raise DeviceError("PyDAQmx import error!")
from ctypes import *

GROUP_DIAGNOSTICS = u"Diagnostics"
GROUP_AI = "Analog inputs"
GROUP_AO = "Analog outputs"
GROUP_DI = "Digital inputs"
GROUP_DO = "Digital outputs"

CONST_AI_TERM_CONF = FrozenOrderedBidict(
    {DAQmx_Val_Bit_TermCfg_RSE: "RSE", DAQmx_Val_Bit_TermCfg_NRSE: "NRSE", DAQmx_Val_Bit_TermCfg_Diff: "Diff",
     DAQmx_Val_Bit_TermCfg_PseudoDIFF: "PseudoDiff"})
CONST_AI_TERM_CONF_INPUT = {DAQmx_Val_Bit_TermCfg_RSE: DAQmx_Val_RSE,
                            DAQmx_Val_Bit_TermCfg_NRSE: DAQmx_Val_NRSE,
                            DAQmx_Val_Bit_TermCfg_Diff: DAQmx_Val_Diff,
                            DAQmx_Val_Bit_TermCfg_PseudoDIFF: DAQmx_Val_PseudoDiff
                            }


class DeviceNiCard(Device):
    task_handle_diagnose = None

    def __init__(self, card_address, device_id=None, config=None):
        self.card_address = card_address
        self.active_tasks = {}
        super(DeviceNiCard, self).__init__(device_id, config)
        self.connect()

    def _init_attributes(self):
        super(DeviceNiCard, self)._init_attributes()
        self.create_attribute(ATTR_SERIAL_NUMBER, readonly=True)
        self.create_attribute(ATTR_PRODUCT_TYPE, readonly=True)
        self.create_attribute(ATTR_ANALOG_INPUT_READOUT_TASK, GROUP_DIAGNOSTICS, default_type=bool,
                              set_function=self.diagnose_analogue_readout_task)
        self.create_attribute(ATTR_TERMINAL_CONFIGURATION, GROUP_DIAGNOSTICS, default_value=u"RSE",
                              default_type=CONST_AI_TERM_CONF.values())
        self.create_attribute(ATTR_SAMPLES_NUMBER, GROUP_DIAGNOSTICS, default_value=2500, default_type=np.uint16,
                              min_value=0, max_value=10000)
        self.create_attribute(ATTR_READOUT_AT_SAMPLES, GROUP_DIAGNOSTICS, default_value=500, default_type=np.uint16,
                              min_value=0, max_value=10000)

    def connect(self, *args):
        """
        Call only this function to connect devices to port / socket / library / ...
        :param args: connect attributes
        """
        try:
            self.connecting = True
            self.get_system_info()
            self.handle_configuration_event()
            self.connected = True
            self.connecting = False
            self.set_status(STATUS_IDLE)
        except DAQError:
            self.handle_connection_error(u"NiCard was not found")
            return

    def handle_configuration(self):
        try:
            start_at = time.time()
            analog_inputs, (min_voltage, max_voltage) = self.get_analog_inputs_list()
            self.attributes[GROUP_AI] = {}
            for analog_input, terminal_config in analog_inputs:
                if analog_input == '': continue
                self.create_attribute(analog_input, GROUP_AI, min_value=min_voltage, max_value=max_voltage,
                                      default_type=np.float64, unit="V", decimals=12, readonly=True,
                                      terminal_config=terminal_config)
            analog_outputs, (min_voltage, max_voltage) = self.get_analog_outputs_list()
            self.attributes[GROUP_AO] = {}
            for analog_output in analog_outputs:
                if analog_output == '': continue
                self.create_attribute(analog_output, GROUP_AO, min_value=min_voltage, max_value=max_voltage,
                                      default_type=np.float64, unit="V", decimals=12, readonly=True)

            digital_outputs = self.get_digital_outputs_list()
            self.attributes[GROUP_DO] = {}
            for digital_output in digital_outputs:
                if digital_output == '': continue
                self.create_attribute(digital_output, GROUP_DO, default_type=bool,
                                      set_function=lambda value, do=digital_output: self.set_digital_output(do, value))

            digital_inputs = self.get_digital_inputs_list()
            self.attributes[GROUP_DI] = {}
            for digital_input in digital_inputs:
                if digital_input == '': continue
                self.create_attribute(digital_input, GROUP_DI, default_type=bool, readonly=True)

            self.get_counters_list()
            self._config_commands()
            self._config_attributes()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))
        except (DAQError, DeviceError):
            self.handle_response_error(u"Could not get device info")
            return

    def command(self, command, callback=None, with_token=False, returning=True):
        raise DeviceError("Device does not accept any commands.")

    def task(self, task_name):
        return Task("{}_{}".format(self.device_id, task_name))

    def diagnose_analogue_readout_task(self, start=True):
        try:
            if start:
                task = self.task("analogue_diagnose_task")
                terminal_configuration = CONST_AI_TERM_CONF.inv[
                    self[GROUP_DIAGNOSTICS][ATTR_TERMINAL_CONFIGURATION][VALUE]]
                channels_in_task = []
                for channel, attribute in self[GROUP_AI].items():
                    if attribute["terminal_config"] < terminal_configuration: continue

                    min_voltage, max_voltage = attribute[MIN], attribute[MAX]
                    task.CreateAIVoltageChan("{}/{}".format(self.card_address, channel), "",
                                             CONST_AI_TERM_CONF_INPUT[terminal_configuration],
                                             min_voltage, max_voltage, DAQmx_Val_Volts, "")
                    channels_in_task.append(channel)
                task.CfgSampClkTiming("", self[GROUP_DIAGNOSTICS][ATTR_SAMPLES_NUMBER][VALUE], DAQmx_Val_Rising,
                                      DAQmx_Val_ContSamps, self[GROUP_DIAGNOSTICS][ATTR_READOUT_AT_SAMPLES][VALUE])

                self.analogueInputCallback_C = DAQmxEveryNSamplesEventCallbackPtr(self.analogueInputCallback)
                task.RegisterEveryNSamplesEvent(DAQmx_Val_Acquired_Into_Buffer,
                                                self[GROUP_DIAGNOSTICS][ATTR_READOUT_AT_SAMPLES][VALUE], 0,
                                                self.analogueInputCallback_C, None)
                task_buffer = np.zeros(self[GROUP_DIAGNOSTICS][ATTR_READOUT_AT_SAMPLES][VALUE] * len(channels_in_task))
                self.add_active_task(task, "analogue_diagnose_task", (task_buffer, channels_in_task))
                self.task_handle_diagnose = task.taskHandle.value
                self.start_task(self.task_handle_diagnose)
            else:
                self.stop_task(self.task_handle_diagnose)
        except DAQError as e:
            raise DeviceError(e)

    def analogueInputCallback(self, taskHandle, status, nSamples, callbackData_ptr):
        try:
            read = int32()
            task, task_handle, (task_buffer, channels_in_task) = self.active_task("analogue_diagnose_task")
            DAQmxReadAnalogF64(taskHandle, self[GROUP_DIAGNOSTICS][ATTR_READOUT_AT_SAMPLES][VALUE], 1.0,
                               DAQmx_Val_GroupByChannel, task_buffer, len(task_buffer), byref(read), None)

            for index, channel in enumerate(channels_in_task):
                buffIndex = index * nSamples
                self.set_raw_value((GROUP_AI, channel), task_buffer[buffIndex:buffIndex + nSamples].mean())
            return 0
        except DAQError as e:
            raise DeviceError(e)

    def get_system_info(self):
        try:
            # Get serial number
            read = uInt32()
            DAQmxGetDevSerialNum(self.card_address, byref(read))
            serial_number = hex(read.value)
            self.set_raw_value(ATTR_SERIAL_NUMBER, serial_number)
            # Get product type
            buffer_size = 100
            read = create_string_buffer(buffer_size)
            DAQmxGetDevProductType(self.card_address, read, buffer_size)
            product_type = read.value.decode()
            self.set_raw_value(ATTR_PRODUCT_TYPE, product_type)

            return product_type, serial_number
        except DAQError as e:
            raise DeviceError(e)

    def get_analog_inputs_list(self):
        try:
            buffer_size = 1000
            read = create_string_buffer(buffer_size)
            DAQmxGetDevAIPhysicalChans(self.card_address, read, buffer_size)
            # There can be many ranges and there is no way to now how many ahead
            voltage_range = (float64 * 256)()
            DAQmxGetDevAIVoltageRngs(self.card_address, voltage_range, 256)

            available_channels = read.value.decode().replace(self.card_address + "/", "").split(", ")

            read = int32()
            channel_pairs = []
            for channel in available_channels:
                DAQmxGetPhysicalChanAITermCfgs("{}/{}".format(self.card_address, channel), byref(read))
                channel_pairs.append((channel, read.value))

            # Return all channels and the biggest range
            return channel_pairs, (min(voltage_range), max(voltage_range))
        except DAQError as e:
            raise DeviceError(u"Error getting analog inputs: {}".format(e))

    def get_analog_outputs_list(self):
        try:
            buffer_size = 1000
            read = create_string_buffer(buffer_size)
            DAQmxGetDevAOPhysicalChans(self.card_address, read, buffer_size)
            # There can be many ranges and there is no way to now how many ahead
            voltage_range = (float64 * 256)()
            DAQmxGetDevAOVoltageRngs(self.card_address, voltage_range, 256)
            # Return all channels and the biggest range
            return read.value.decode().replace(self.card_address + "/", "").split(", "), (
                min(voltage_range), max(voltage_range))
        except DAQError as e:
            raise DeviceError(u"Error getting analog outputs: {}".format(e))

    def get_digital_inputs_list(self):
        try:
            buffer_size = 1000
            read = create_string_buffer(buffer_size)
            DAQmxGetDevDILines(self.card_address, read, buffer_size)
            return read.value.decode().replace(self.card_address + "/", "").split(", ")
        except DAQError:
            raise DeviceError(u"Error getting digital inputs")

    def get_digital_outputs_list(self):
        buffer_size = 1000
        read = create_string_buffer(buffer_size)
        DAQmxGetDevDOLines(self.card_address, read, buffer_size)
        return read.value.decode().replace(self.card_address + "/", "").split(", ")

    def get_counters_list(self):
        buffer_size = 1000
        read = create_string_buffer(buffer_size)
        DAQmxGetDevCOPhysicalChans(self.card_address, read, buffer_size)
        return read.value.decode().replace(self.card_address + "/", "").split(", ")

    def read_digital_input(self):
        try:
            self.read_di_task = self.task("get_digital_output_task")
            self.add_active_task(self.read_di_task, "get_digital_output_task")
            self.read_di_task.CreateDIChan("{}/port0/line0:7".format(self.card_address), "", DAQmx_Val_ChanForAllLines)
            self.read_di_task.CreateDIChan("{}/port1/line0:7".format(self.card_address), "",
                                           DAQmx_Val_ChanForAllLines)
            self.read_di_task.CreateDIChan("{}/port2/line0:7".format(self.card_address), "",
                                           DAQmx_Val_ChanForAllLines)
            self.start_task(self.read_di_task.taskHandle.value)
            task_buffer = np.zeros([100], dtype=uInt8)
            read = int32()
            bytesPerSamp = int32()
            DAQmxReadDigitalLines(self.read_di_task.taskHandle.value, 1, 10.0, DAQmx_Val_GroupByChannel, task_buffer,
                                  100, byref(read), byref(bytesPerSamp), None)
            for i, di_line in enumerate(self[GROUP_DI].keys()):
                self.set_raw_value([GROUP_DI, di_line], task_buffer[i])
            self.stop_task("get_digital_output_task")
        except DAQError:
            raise DeviceError(u"Error getting digital outputs")

    def read_digital_output(self):
        try:
            self.read_do_task = self.task("get_digital_input_task")
            self.add_active_task(self.read_do_task, "get_digital_input_task")
            self.read_do_task.CreateDOChan("{}/port0/line0:7".format(self.card_address), "", DAQmx_Val_ChanForAllLines)
            self.read_do_task.CreateDOChan("{}/port1/line0:7".format(self.card_address), "",
                                           DAQmx_Val_ChanForAllLines)
            self.read_do_task.CreateDOChan("{}/port2/line0:7".format(self.card_address), "",
                                           DAQmx_Val_ChanForAllLines)
            self.start_task(self.read_do_task.taskHandle.value)
            task_buffer = np.zeros([1000], dtype=uInt8)
            read = int32()
            bytesPerSamp = int32()
            DAQmxReadDigitalLines(self.read_do_task.taskHandle.value, 1, 10.0, DAQmx_Val_GroupByChannel, task_buffer,
                                  100, byref(read), byref(bytesPerSamp), None)
            for i, do_line in enumerate(self[GROUP_DO].keys()):
                self.set_raw_value([GROUP_DO, do_line], task_buffer[i])
            self.stop_task("get_digital_input_task")
        except DAQError:
            raise DeviceError(u"Error getting digital inputs")

    def set_digital_output(self, line, value):
        try:
            self.write_do_task = self.task("set_digital_output_task")
            self.add_active_task(self.write_do_task, "set_digital_output_task")
            self.write_do_task.CreateDOChan("{}/{}".format(self.card_address, line), "", DAQmx_Val_ChanForAllLines)
            self.start_task(self.write_do_task.taskHandle.value)
            data = np.ndarray([1], dtype=uInt8)
            data[0] = int(value)
            DAQmxWriteDigitalLines(self.write_do_task.taskHandle.value, 1, 1, 10.0, DAQmx_Val_GroupByChannel, data,
                                   None, None)
            self.stop_task("set_digital_output_task")
        except DAQError as e:
            print(e)
            raise DeviceError(u"Error setting digital output")

    def stop_task(self, task_name):
        self.logger.debug(u"Clearing task {}".format(task_name))
        try:
            self.active_tasks[task_name][0].ClearTask()
            del self.active_tasks[task_name]
        except DAQError as e:
            raise DeviceError(e)

    def start_task(self, task_handle):
        self.logger.debug(u"Starting task {}".format(task_handle))
        try:
            DAQmxStartTask(task_handle)
        except DAQError as e:
            raise DeviceError(e)

    @expose_method()
    def stop_all_tasks(self):
        for task_handle in list(self.active_tasks.keys())[:]:
            self.stop_task(task_handle)

    def add_active_task(self, task, task_name, task_resources=None):
        self.active_tasks[task_name] = (task, task.taskHandle.value, task_resources)

    def active_task(self, task_name):
        return self.active_tasks.get(task_name, None)

    def command(self, command, callback=None, with_token=False):
        raise DeviceError(u"Not accepting commands")

    def close(self):
        """
        Handle all devices closing stuff here.
        Close socket, close port, etc...        `
        """
        if self.closing:
            return False
        else:
            self.closing = True
            self.connected = False
            self.set_status(STATUS_DISCONNECTING)
            self.stop_all_tasks()
            self.set_status(STATUS_DISCONNECTED)
            self.closing = False
            self.connected = False



import argparse
import os
import signal

import oyaml as yaml

import kamzik3
from kamzik3.snippets.snippetsWidgets import init_qt_app

# Parse arguments from commandline
parser = argparse.ArgumentParser()
parser.add_argument("--chdir", help="Path to active directory",
                    default="./")
parser.add_argument("--conf", help="Path to configuration file in yaml format",
                    default="./examples/conf_dummy_client.yml")
args = parser.parse_args()

# Set active directory
os.chdir(args.chdir)
# Create PyQT5 application
app = init_qt_app(enable_hd_scaling=False)

# Load configuration file
with open(args.conf, "r") as configFile:
    config = yaml.load(configFile, Loader=yaml.Loader)

signal.signal(signal.SIGINT, signal.SIG_DFL)
if "session_window" in config:
    config["session_window"].show()

# Start PyQT5 app loop
app.exec_()

# Close session
kamzik3.session.stop()



import numpy as np

from kamzik3 import MacroException
from kamzik3 import session
from kamzik3.constants import *
from kamzik3.devices.general.deviceScanner import DeviceScanner
from kamzik3.macro.macro import Macro
from kamzik3.macro.step import StepDeviceAttributeNumerical
from kamzik3.snippets.snippetsDecorators import expose_method


class GeneralFlyScanner(DeviceScanner):
    attribute_filter = [ATTR_VELOCITY, ATTR_POSITION]

    def _init_attributes(self):
        DeviceScanner._init_attributes(self)
        self.create_attribute(ATTR_DWELL_TIME, default_value=100, unit="ms", min_value=0, default_type=np.int64)

    @expose_method()
    def get_scanner_attributes(self):
        return [ATTR_DWELL_TIME]

    @expose_method()
    def get_scanner_macro(self, scanner_input, scanner_attributes):
        device_id = scanner_input.step_attributes['device_id']
        device = session.get_device(device_id)
        macro = Macro(device_id="{}_{}".format(self.device_id, scanner_input.common_id))
        distance = abs(scanner_input.start_point - scanner_input.end_point)
        dwell_time = scanner_attributes[ATTR_DWELL_TIME]
        velocity = device.get_attribute(ATTR_VELOCITY)
        maximum_velocity = velocity.maximum()
        scan_line_time = dwell_time * scanner_input.steps_count
        final_velocity = distance / scan_line_time
        final_velocity = final_velocity.to(maximum_velocity.u)

        if final_velocity > maximum_velocity:
            raise MacroException(
                "Calculated velocity {} is higher than maximum velocity {}.".format(final_velocity, maximum_velocity))

        scanner_input.set_steps_count(1)
        max_velocity_step = StepDeviceAttributeNumerical("SetMaxVelocityStep", device_id, ATTR_VELOCITY,
                                                         maximum_velocity,
                                                         trigger_log=False,
                                                         negative_tolerance=velocity.negative_tolerance(),
                                                         positive_tolerance=velocity.positive_tolerance())
        if not scanner_input.bidirectional:
            macro.add(max_velocity_step)
            macro.add(
                StepDeviceAttributeNumerical("SetInitPositionStep", device_id, ATTR_POSITION, scanner_input.start_point,
                                             trigger_log=False,
                                             negative_tolerance=scanner_input.step_attributes["negative_tolerance"],
                                             positive_tolerance=scanner_input.step_attributes["positive_tolerance"]))

        macro.add(
            StepDeviceAttributeNumerical("FinalVelocityStep", device_id, ATTR_VELOCITY, final_velocity,
                                         trigger_log=False, negative_tolerance=velocity.negative_tolerance(),
                                         positive_tolerance=velocity.positive_tolerance()))

        previous_reset_step = scanner_input.get_reset_step()

        def get_reset_step():
            return [max_velocity_step, previous_reset_step]

        scanner_input.get_reset_step = get_reset_step
        macro.add(scanner_input)
        return macro



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'deviceDebugTemplate.ui'
#
# Created by: PyQt5 UI code generator 5.10.1
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(724, 222)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        self.widgets_stack = QtWidgets.QStackedWidget(Form)
        self.widgets_stack.setObjectName("widgets_stack")
        self.page = QtWidgets.QWidget()
        self.page.setObjectName("page")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.page)
        self.gridLayout_2.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.widget_attributes = QtWidgets.QWidget(self.page)
        self.widget_attributes.setObjectName("widget_attributes")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.widget_attributes)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setObjectName("verticalLayout")
        self.tab_attribute_groups = QtWidgets.QTabWidget(self.widget_attributes)
        self.tab_attribute_groups.setObjectName("tab_attribute_groups")
        self.verticalLayout.addWidget(self.tab_attribute_groups)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setSpacing(6)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.input_attribute_filter = QtWidgets.QLineEdit(self.widget_attributes)
        self.input_attribute_filter.setInputMask("")
        self.input_attribute_filter.setObjectName("input_attribute_filter")
        self.horizontalLayout_2.addWidget(self.input_attribute_filter)
        self.pushButton_2 = QtWidgets.QPushButton(self.widget_attributes)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/icons/search.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.pushButton_2.setIcon(icon)
        self.pushButton_2.setObjectName("pushButton_2")
        self.horizontalLayout_2.addWidget(self.pushButton_2)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.gridLayout_2.addWidget(self.widget_attributes, 0, 0, 1, 1)
        self.widgets_stack.addWidget(self.page)
        self.page_2 = QtWidgets.QWidget()
        self.page_2.setObjectName("page_2")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.page_2)
        self.gridLayout_3.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_3.setObjectName("gridLayout_3")
        self.widget_configuration = QtWidgets.QWidget(self.page_2)
        self.widget_configuration.setObjectName("widget_configuration")
        self.gridLayout_3.addWidget(self.widget_configuration, 0, 0, 1, 1)
        self.widgets_stack.addWidget(self.page_2)
        self.page_3 = QtWidgets.QWidget()
        self.page_3.setObjectName("page_3")
        self.gridLayout_4 = QtWidgets.QGridLayout(self.page_3)
        self.gridLayout_4.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_4.setObjectName("gridLayout_4")
        self.widget_terminal = QtWidgets.QWidget(self.page_3)
        self.widget_terminal.setObjectName("widget_terminal")
        self.gridLayout_5 = QtWidgets.QGridLayout(self.widget_terminal)
        self.gridLayout_5.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_5.setObjectName("gridLayout_5")
        self.label_terminator = QtWidgets.QLabel(self.widget_terminal)
        self.label_terminator.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_terminator.setObjectName("label_terminator")
        self.gridLayout_5.addWidget(self.label_terminator, 0, 2, 1, 1)
        self.label_command = QtWidgets.QLabel(self.widget_terminal)
        self.label_command.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_command.setObjectName("label_command")
        self.gridLayout_5.addWidget(self.label_command, 0, 0, 1, 1)
        self.input_command = QtWidgets.QLineEdit(self.widget_terminal)
        self.input_command.setObjectName("input_command")
        self.gridLayout_5.addWidget(self.input_command, 0, 1, 1, 1)
        self.button_send = QtWidgets.QPushButton(self.widget_terminal)
        self.button_send.setObjectName("button_send")
        self.gridLayout_5.addWidget(self.button_send, 0, 4, 1, 1)
        self.output_commands = QtWidgets.QTextBrowser(self.widget_terminal)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.output_commands.sizePolicy().hasHeightForWidth())
        self.output_commands.setSizePolicy(sizePolicy)
        self.output_commands.setMinimumSize(QtCore.QSize(0, 50))
        self.output_commands.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.output_commands.setObjectName("output_commands")
        self.gridLayout_5.addWidget(self.output_commands, 1, 0, 1, 5)
        self.comboBox_terminator = QtWidgets.QComboBox(self.widget_terminal)
        self.comboBox_terminator.setObjectName("comboBox_terminator")
        self.comboBox_terminator.addItem("")
        self.comboBox_terminator.addItem("")
        self.comboBox_terminator.addItem("")
        self.comboBox_terminator.addItem("")
        self.comboBox_terminator.addItem("")
        self.gridLayout_5.addWidget(self.comboBox_terminator, 0, 3, 1, 1)
        self.gridLayout_4.addWidget(self.widget_terminal, 0, 0, 1, 1)
        self.widgets_stack.addWidget(self.page_3)
        self.page_4 = QtWidgets.QWidget()
        self.page_4.setObjectName("page_4")
        self.gridLayout_6 = QtWidgets.QGridLayout(self.page_4)
        self.gridLayout_6.setContentsMargins(0, 0, 0, 0)
        self.gridLayout_6.setVerticalSpacing(6)
        self.gridLayout_6.setObjectName("gridLayout_6")
        self.scroll_area_methods = QtWidgets.QScrollArea(self.page_4)
        self.scroll_area_methods.setWidgetResizable(True)
        self.scroll_area_methods.setObjectName("scroll_area_methods")
        self.widget_grid_for_methods = QtWidgets.QWidget()
        self.widget_grid_for_methods.setGeometry(QtCore.QRect(0, 0, 98, 28))
        self.widget_grid_for_methods.setObjectName("widget_grid_for_methods")
        self.gridLayout_7 = QtWidgets.QGridLayout(self.widget_grid_for_methods)
        self.gridLayout_7.setObjectName("gridLayout_7")
        self.scroll_area_methods.setWidget(self.widget_grid_for_methods)
        self.gridLayout_6.addWidget(self.scroll_area_methods, 0, 0, 1, 1)
        self.output_methods = QtWidgets.QPlainTextEdit(self.page_4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.output_methods.sizePolicy().hasHeightForWidth())
        self.output_methods.setSizePolicy(sizePolicy)
        self.output_methods.setMaximumSize(QtCore.QSize(16777215, 50))
        self.output_methods.setReadOnly(True)
        self.output_methods.setOverwriteMode(True)
        self.output_methods.setObjectName("output_methods")
        self.gridLayout_6.addWidget(self.output_methods, 1, 0, 1, 1)
        self.widgets_stack.addWidget(self.page_4)
        self.gridLayout.addWidget(self.widgets_stack, 2, 0, 1, 1)
        self.label_device_id = QtWidgets.QLabel(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_device_id.sizePolicy().hasHeightForWidth())
        self.label_device_id.setSizePolicy(sizePolicy)
        self.label_device_id.setMinimumSize(QtCore.QSize(0, 35))
        font = QtGui.QFont()
        font.setPointSize(23)
        font.setBold(True)
        font.setWeight(75)
        self.label_device_id.setFont(font)
        self.label_device_id.setStyleSheet("QLabel{background:#68d532;}")
        self.label_device_id.setAlignment(QtCore.Qt.AlignCenter)
        self.label_device_id.setObjectName("label_device_id")
        self.gridLayout.addWidget(self.label_device_id, 0, 0, 1, 1)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.button_attributes = QtWidgets.QPushButton(Form)
        self.button_attributes.setCursor(QtGui.QCursor(QtCore.Qt.OpenHandCursor))
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/icons/icons/settings.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_attributes.setIcon(icon1)
        self.button_attributes.setIconSize(QtCore.QSize(20, 20))
        self.button_attributes.setCheckable(True)
        self.button_attributes.setChecked(True)
        self.button_attributes.setAutoExclusive(True)
        self.button_attributes.setObjectName("button_attributes")
        self.horizontalLayout.addWidget(self.button_attributes)
        self.button_attributes_2 = QtWidgets.QPushButton(Form)
        self.button_attributes_2.setCursor(QtGui.QCursor(QtCore.Qt.OpenHandCursor))
        self.button_attributes_2.setIcon(icon1)
        self.button_attributes_2.setIconSize(QtCore.QSize(20, 20))
        self.button_attributes_2.setCheckable(True)
        self.button_attributes_2.setChecked(False)
        self.button_attributes_2.setAutoExclusive(True)
        self.button_attributes_2.setObjectName("button_attributes_2")
        self.horizontalLayout.addWidget(self.button_attributes_2)
        self.button_configuration = QtWidgets.QPushButton(Form)
        self.button_configuration.setCursor(QtGui.QCursor(QtCore.Qt.OpenHandCursor))
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/icons/icons/configFile.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_configuration.setIcon(icon2)
        self.button_configuration.setIconSize(QtCore.QSize(20, 20))
        self.button_configuration.setCheckable(True)
        self.button_configuration.setAutoExclusive(True)
        self.button_configuration.setObjectName("button_configuration")
        self.horizontalLayout.addWidget(self.button_configuration)
        self.button_terminal = QtWidgets.QPushButton(Form)
        self.button_terminal.setCursor(QtGui.QCursor(QtCore.Qt.OpenHandCursor))
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/icons/icons/terminal.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_terminal.setIcon(icon3)
        self.button_terminal.setIconSize(QtCore.QSize(20, 20))
        self.button_terminal.setCheckable(True)
        self.button_terminal.setAutoExclusive(True)
        self.button_terminal.setObjectName("button_terminal")
        self.horizontalLayout.addWidget(self.button_terminal)
        spacerItem = QtWidgets.QSpacerItem(20, 0, QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.button_enable = QtWidgets.QPushButton(Form)
        self.button_enable.setCursor(QtGui.QCursor(QtCore.Qt.OpenHandCursor))
        icon4 = QtGui.QIcon()
        icon4.addPixmap(QtGui.QPixmap(":/icons/icons/play.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_enable.setIcon(icon4)
        self.button_enable.setIconSize(QtCore.QSize(20, 20))
        self.button_enable.setCheckable(False)
        self.button_enable.setAutoExclusive(False)
        self.button_enable.setObjectName("button_enable")
        self.horizontalLayout.addWidget(self.button_enable)
        self.button_disable = QtWidgets.QPushButton(Form)
        self.button_disable.setCursor(QtGui.QCursor(QtCore.Qt.OpenHandCursor))
        icon5 = QtGui.QIcon()
        icon5.addPixmap(QtGui.QPixmap(":/icons/icons/cancel.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_disable.setIcon(icon5)
        self.button_disable.setIconSize(QtCore.QSize(20, 20))
        self.button_disable.setCheckable(False)
        self.button_disable.setAutoExclusive(False)
        self.button_disable.setObjectName("button_disable")
        self.horizontalLayout.addWidget(self.button_disable)
        self.button_connect = QtWidgets.QPushButton(Form)
        self.button_connect.setCursor(QtGui.QCursor(QtCore.Qt.OpenHandCursor))
        icon6 = QtGui.QIcon()
        icon6.addPixmap(QtGui.QPixmap(":/icons/icons/plug.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_connect.setIcon(icon6)
        self.button_connect.setIconSize(QtCore.QSize(20, 20))
        self.button_connect.setCheckable(False)
        self.button_connect.setAutoExclusive(False)
        self.button_connect.setObjectName("button_connect")
        self.horizontalLayout.addWidget(self.button_connect)
        self.button_disconnect = QtWidgets.QPushButton(Form)
        self.button_disconnect.setCursor(QtGui.QCursor(QtCore.Qt.OpenHandCursor))
        icon7 = QtGui.QIcon()
        icon7.addPixmap(QtGui.QPixmap(":/icons/icons/logout.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_disconnect.setIcon(icon7)
        self.button_disconnect.setIconSize(QtCore.QSize(20, 20))
        self.button_disconnect.setCheckable(False)
        self.button_disconnect.setAutoExclusive(False)
        self.button_disconnect.setObjectName("button_disconnect")
        self.horizontalLayout.addWidget(self.button_disconnect)
        self.gridLayout.addLayout(self.horizontalLayout, 1, 0, 1, 1)

        self.retranslateUi(Form)
        self.widgets_stack.setCurrentIndex(0)
        self.button_attributes.clicked.connect(Form.show_attributes_widget)
        self.button_configuration.clicked.connect(Form.show_configuration_widget)
        self.button_terminal.clicked.connect(Form.show_terminal_widget)
        self.button_send.clicked.connect(Form.send_command)
        self.input_command.returnPressed.connect(Form.send_command)
        self.button_disconnect.clicked.connect(Form.disconnect_device)
        self.button_connect.clicked.connect(Form.connect_device)
        self.button_disable.clicked.connect(Form.disable_device)
        self.button_enable.clicked.connect(Form.enable_device)
        self.tab_attribute_groups.currentChanged['int'].connect(Form.slot_attribute_group_changed)
        self.button_attributes_2.clicked.connect(Form.show_methods_widget)
        self.comboBox_terminator.activated['QString'].connect(Form.set_command_terminator)
        self.pushButton_2.clicked.connect(Form.filter_attributes)
        self.input_attribute_filter.returnPressed.connect(self.pushButton_2.click)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.input_attribute_filter.setPlaceholderText(_translate("Form", "Filter attributes"))
        self.pushButton_2.setText(_translate("Form", "Filter"))
        self.label_terminator.setText(_translate("Form", "Terminator:"))
        self.label_command.setText(_translate("Form", "Command:"))
        self.button_send.setText(_translate("Form", "Send"))
        self.comboBox_terminator.setItemText(0, _translate("Form", "None"))
        self.comboBox_terminator.setItemText(1, _translate("Form", "\\r\\n"))
        self.comboBox_terminator.setItemText(2, _translate("Form", "\\r"))
        self.comboBox_terminator.setItemText(3, _translate("Form", "\\n"))
        self.comboBox_terminator.setItemText(4, _translate("Form", "\\t"))
        self.label_device_id.setText(_translate("Form", "Device Id"))
        self.button_attributes.setToolTip(_translate("Form", "Show device attributes page"))
        self.button_attributes.setText(_translate("Form", "Attributes"))
        self.button_attributes_2.setToolTip(_translate("Form", "Show device attributes page"))
        self.button_attributes_2.setText(_translate("Form", "Methods"))
        self.button_configuration.setToolTip(_translate("Form", "Show device configuration page"))
        self.button_configuration.setText(_translate("Form", "Configuration"))
        self.button_terminal.setToolTip(_translate("Form", "Show device terminal page"))
        self.button_terminal.setText(_translate("Form", "Terminal"))
        self.button_enable.setToolTip(_translate("Form", "Enable device"))
        self.button_enable.setText(_translate("Form", "Enable"))
        self.button_disable.setToolTip(_translate("Form", "Disable device"))
        self.button_disable.setText(_translate("Form", "Disable"))
        self.button_connect.setToolTip(_translate("Form", "Connect device"))
        self.button_connect.setText(_translate("Form", "Connect"))
        self.button_disconnect.setToolTip(_translate("Form", "Disconnect device"))
        self.button_disconnect.setText(_translate("Form", "Disconnect"))

from kamzik3 import resource_kamzik3_rc



import os
import time

from kamzik3 import DeviceError
from kamzik3.constants import *
from kamzik3.devices.device import Device


class DeviceFileSync(Device):

    def __init__(self, filepath, device_id=None, config=None):
        self.filepath = filepath
        super(DeviceFileSync, self).__init__(device_id, config)
        self.connect()

    def _init_attributes(self):
        super(DeviceFileSync, self)._init_attributes()
        self.create_attribute(ATTR_FILEPATH, set_function=self.set_filepath, readonly=True)
        self.create_attribute(ATTR_CONTENT, set_function=self.set_content)

    def handle_configuration(self):

        start_at = time.time()

        self._config_commands()
        self._config_attributes()
        self.set_value(ATTR_FILEPATH, self.filepath)
        self.set_status(STATUS_CONFIGURED)
        self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

    def command(self, command, callback=None, with_token=False, returning=True):
        raise DeviceError("Device does not accept any commands.")

    def set_filepath(self, value):
        if not os.path.exists(value):
            raise DeviceError("File {} does not exists".format(value))
        else:
            with open(value, "r") as fp:
                file_content = "".join(fp.readlines())
                self.set_raw_value(ATTR_CONTENT, file_content)

    def set_content(self, value):
        with open(self.get_value(ATTR_FILEPATH), "w") as fp:
            fp.write(value)



import copy

from PyQt5 import QtCore, QtGui
from PyQt5.QtCore import QItemSelection, pyqtSlot, Qt, QItemSelectionModel, QAbstractTableModel
from PyQt5.QtWidgets import QWidget
from kamzik3.constants import *
from kamzik3.gui.templates.modelTableTemplate import Ui_Form

Qt.CopyRole = Qt.UserRole + 1

class ModelTableModel(QAbstractTableModel):

    def __init__(self, model_data, empty_row ,parent=None):
        assert ROWS in model_data
        assert COLUMNS in model_data
        self.model_data = model_data
        self.empty_row = empty_row
        QAbstractTableModel.__init__(self, parent)

    def columnCount(self, parent=None, *args, **kwargs):
        return len(self.model_data[COLUMNS])

    def rowCount(self, parent=None, *args, **kwargs):
        return len(self.model_data[ROWS])

    def headerData(self, p_int, Qt_Orientation, role=None):
        if role == Qt.DisplayRole and Qt_Orientation == Qt.Horizontal:
            return self.model_data[COLUMNS][p_int]
        elif role == Qt.DisplayRole and Qt_Orientation == Qt.Vertical:
            return p_int + 1

    def data(self, index, role):
        if role in (Qt.DisplayRole, Qt.EditRole):
            return self.model_data[ROWS][index.row()][index.column()]

    def setData(self, index, value, role):
        if role == Qt.EditRole:
            self.model_data[ROWS][index.row()][index.column()] = value
        elif role == Qt.CopyRole:
            self.model_data[ROWS][index] = copy.deepcopy(self.model_data[ROWS][value])
            index = self.index(index, 0)
        self.dataChanged.emit(index, index, [role])
        return True

    def insertRows(self, row, count, parent):
        self.beginInsertRows(parent, row, row + count - 1)
        for i in range(count):
            self.model_data[ROWS].insert(row + i, copy.deepcopy(self.empty_row))
        self.endInsertRows()
        return True

    def removeRows(self, row, count, parent):
        self.beginRemoveRows(parent, row, row + count - 1)
        for i in range(count):
            del self.model_data[ROWS][row + i]
        self.endRemoveRows()
        return True

    def moveRows(self, sourceParent, sourceRow, count, destinationParent, destinationChild):
        self.beginMoveRows(sourceParent, sourceRow, sourceRow + count - 1, destinationParent, destinationChild)
        for i in range(count):
            self.model_data[ROWS].insert(destinationChild + i, self.model_data[ROWS].pop(sourceRow + i))
        self.endMoveRows()
        return True

    def setHeaderData(self, section, orientation, data, role):
        if role == Qt.EditRole and orientation == Qt.Horizontal:
            self.model_data[COLUMNS][section] = data
        self.headerDataChanged.emit(orientation, section, section)

    def flags(self, index):
        return QtCore.Qt.ItemIsEditable | QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable

class ModelTable(QWidget, Ui_Form):

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        self.setupUi(self)

    def setupUi(self, Form):
        super().setupUi(Form)
        self.slot_selection_changed(QItemSelection(), QItemSelection())

    def set_model(self, model):
        assert isinstance(model, ModelTableModel)
        self.input_table_view.setModel(model)
        self.input_table_view.selectionModel().selectionChanged.disconnect()
        self.input_table_view.selectionModel().selectionChanged.connect(self.slot_selection_changed)

    def get_model(self):
        return self.input_table_view.model()

    @pyqtSlot(QItemSelection, QItemSelection)
    def slot_selection_changed(self, selected, deselected):
        selected = selected.indexes()
        for w in (self.button_copy_item, self.button_move_item_top, self.button_move_item_up,
                  self.button_remove_item, self.button_move_item_down, self.button_move_item_bottom):
            w.setDisabled(len(selected) == 0)
        if len(selected) == 0:
            return

        if selected[-1].row() == self.input_table_view.model().rowCount() - 1:
            for w in (self.button_move_item_down, self.button_move_item_bottom):
                w.setDisabled(True)
        if selected[-1].row() == 0:
            for w in (self.button_move_item_up, self.button_move_item_top):
                w.setDisabled(True)

    @pyqtSlot()
    def slot_add_item(self):
        selection = self.input_table_view.selectionModel().selectedIndexes()
        row = 0 if len(selection) == 0 else selection[-1].row() + 1
        data_model = self.input_table_view.model()
        data_model.insertRow(row)
        self._select_item_row(row)
        return row

    @pyqtSlot()
    def slot_copy_item(self):
        new_row = self.slot_add_item()
        model = self.get_model()
        model.setData(new_row, new_row - 1, Qt.CopyRole)

    @pyqtSlot()
    def slot_remove_item(self):
        selection = self.input_table_view.selectionModel().selectedIndexes()
        row = selection[-1].row()
        data_model = self.input_table_view.model()
        data_model.removeRow(row)

        # Select first column of added row
        row_count = self.input_table_view.model().rowCount()
        if row_count <= row:
            row = row_count - 1
        self._select_item_row(row)
        return row

    @pyqtSlot()
    def slot_move_item_top(self):
        selection = self.input_table_view.selectionModel().selectedIndexes()
        row = selection[-1].row()
        data_model = self.input_table_view.model()
        data_model.moveRow(selection[-1], row, self.input_table_view.model().index(0, 0), 0)
        self._select_item_row(0)
        return 0

    @pyqtSlot()
    def slot_move_item_up(self):
        selection = self.input_table_view.selectionModel().selectedIndexes()
        row = selection[-1].row()
        data_model = self.input_table_view.model()
        data_model.moveRow(selection[-1], row, self.input_table_view.model().index(row - 1, 0), row - 1)
        self._select_item_row(row - 1)
        return row - 1

    @pyqtSlot()
    def slot_move_item_down(self):
        selection = self.input_table_view.selectionModel().selectedIndexes()
        row = selection[-1].row()
        data_model = self.input_table_view.model()
        data_model.moveRow(selection[-1], row, self.input_table_view.model().index(row + 1, 0), row + 1)
        self._select_item_row(row + 1)
        return row + 1

    @pyqtSlot()
    def slot_move_item_bottom(self):
        selection = self.input_table_view.selectionModel().selectedIndexes()
        row = selection[-1].row()
        data_model = self.input_table_view.model()
        row_count = data_model.rowCount()
        data_model.moveRow(selection[-1], row, self.input_table_view.model().index(row_count, 0), row_count)
        self._select_item_row(row_count - 1)
        return row_count - 1

    def _select_item_row(self, row):
        selection = self.input_table_view.model().index(row, 0)
        self.input_table_view.selectionModel().clearSelection()
        self.input_table_view.selectionModel().setCurrentIndex(selection, QItemSelectionModel.Select)
        self.input_table_view.setFocus()

    def hideEvent(self, event):
        self.input_table_view.clearSelection()
        self.slot_selection_changed(QItemSelection(), QItemSelection())




import logging
import time
from io import StringIO

import oyaml as yaml
from PyQt5.QtCore import QAbstractTableModel, Qt, pyqtSlot, QModelIndex, pyqtSignal, QItemSelection
from PyQt5.QtGui import QBrush, QColor
from PyQt5.QtWidgets import QWidget, QTableView, QHeaderView, QFileDialog, QColorDialog
from pint import DimensionalityError

import kamzik3
from kamzik3 import DeviceError, DeviceUnknownError, units
from kamzik3.constants import *
from kamzik3.gui.attributeDeviceDisplayWidget import AttributeDeviceDisplayWidget
from kamzik3.gui.deviceWidget import DeviceWidget
from kamzik3.gui.templates.savedAttributesTemplate import Ui_Form
from kamzik3.snippets.snippetsUnits import device_units
from kamzik3.snippets.snippetsWidgets import show_error_message, show_question_dialog, show_prompt_dialog

# Example of file format
"""
Stages:
  columns:
  - !!python/tuple
    - Comment
    - ['', '']
  - !!python/tuple
    - Smaract_axis_0
    - [Position, Value]
  - !!python/tuple
    - Smaract_axis_1
    - [Position, Value]
  - !!python/tuple
    - Smaract_axis_2
    - [Position, Value]
  rows: []
"""


class ViewModel(QAbstractTableModel):

    def __init__(self, model_data, parent=None):
        self.model_data = model_data
        QAbstractTableModel.__init__(self, parent)

    def columnCount(self, parent=None, *args, **kwargs):
        columns = self.model_data.get("columns")
        if columns is None:
            return 0
        return len(columns)

    def rowCount(self, parent=None, *args, **kwargs):
        rows = self.model_data.get("rows")
        if rows is None:
            return 0
        return len(rows)

    def data(self, index, role=None):
        row, column = index.row(), index.column()

        if role == Qt.DisplayRole:
            return "{}".format(self.model_data["rows"][row][column + 2])
        elif role == Qt.BackgroundRole:
            return QBrush(QColor(self.model_data["rows"][row][0]))
        elif role == Qt.EditRole:
            current_value = self.model_data["rows"][index.row()][2]
            return current_value

    def headerData(self, p_int, Qt_Orientation, role=None):
        if role == Qt.DisplayRole and Qt_Orientation == Qt.Horizontal:
            data = self.model_data["columns"][p_int]
            header_text = "{} {}".format(data[0], data[1][-2])
            return header_text
        elif role == Qt.DisplayRole and Qt_Orientation == Qt.Vertical:
            return "{}".format(self.model_data["rows"][p_int][1])

    def setData(self, index, value, role=None):

        if role == Qt.EditRole and value != "":
            self.model_data["rows"][index.row()][2] = value
            self.dataChanged.emit(index, index)
            return True

        return False

    def flags(self, index):
        if index.column() == 0:
            return Qt.ItemIsSelectable | Qt.ItemIsEditable | Qt.ItemIsEnabled

        return Qt.ItemIsSelectable | Qt.ItemIsEnabled


class SavedAttributesWidget(Ui_Form, DeviceWidget):
    default_row_color = "#ffffff"
    file_path = None
    sigUpdateData = pyqtSignal("PyQt_PyObject")

    def __init__(self, device=None, config=None, parent=None):
        self.current_data = {}
        self.preset_groups = {}
        self.current_preset_group = "Default"
        self.rows_selection_mode = False
        self.logger = logging.getLogger("Gui.Device.SavedAttributeWidget")

        if device is None:
            QWidget.__init__(self, parent)
        else:
            DeviceWidget.__init__(self, device, config=config, parent=parent)


        self.session = kamzik3.session
        self.manipulation_buttons = [self.button_set_value, self.button_change_color, self.button_move_top,
                                     self.button_move_up, self.button_move_down, self.button_move_bottom,
                                     self.button_remove]
        for button in self.manipulation_buttons:
            button.setDisabled(True)

        self.preset_group_buttons = (self.combo_preset, self.button_remove_preset_group, self.button_add_preset_group,
                                     self.button_add_preset_group_from_selected_rows)
        self.preset_rows_selection_buttons = (self.button_confirm_rows_selection, self.button_cancel_rows_selection)
        self.sigUpdateData.connect(self.update_data)
        self.tab_groups.currentChanged.connect(self.current_tab_changed)

        for button in self.preset_rows_selection_buttons:
            button.hide()

    def set_source_file(self, file_path):
        self.file_path = file_path
        with open(file_path, "r") as fp:
            self.current_data = yaml.load(fp, Loader=yaml.Loader)
            self.load_data(self.current_data)

    def set_source_file_device(self, file_sync_device):
        DeviceWidget.__init__(self, file_sync_device, self.parent(), setup_ui=False)
        self.file_path = None

    @pyqtSlot("PyQt_PyObject")
    def slot_set_status(self, value):
        if value in READY_DEVICE_STATUSES:
            self.slot_handle_configuration()
            file_content = self.device.get_value(ATTR_CONTENT)
            if file_content is not None:
                file_content = "".join(file_content)
                self.load_data(yaml.load(StringIO(file_content), Loader=yaml.Loader))

            self.device.attach_attribute_callback(ATTR_CONTENT, self.content_update)
            self.checkbox_synced_with_server.setChecked(True)
        elif value in (STATUS_DISCONNECTED, STATUS_DISCONNECTING):
            self.checkbox_synced_with_server.setChecked(False)

    def slot_handle_configuration(self):
        pass

    @pyqtSlot()
    def update_source(self):
        if self.device is not None:
            self.device.set_attribute((ATTR_CONTENT, VALUE), yaml.dump(self.preset_groups))
        elif self.file_path is not None:
            with open(self.file_path, "w") as fp:
                yaml.dump(self.preset_groups, fp, default_flow_style=False)

    @pyqtSlot()
    def content_update(self, key, value):
        if key == VALUE:
            self.sigUpdateData.emit(yaml.load(StringIO(value), Loader=yaml.Loader))

    @pyqtSlot("PyQt_PyObject")
    def load_data(self, data):
        while self.tab_groups.count() != 0:
            self.tab_groups.removeTab(0)

        self.preset_groups = data
        self.current_data = data["Preset groups"][self.current_preset_group]
        self.update_preset_groups(block_signals=True)

        for group_title, model_data in self.current_data.items():
            self.add_group(group_title, model_data)

    @pyqtSlot("PyQt_PyObject")
    def update_data(self, data):
        self.preset_groups = data
        self.current_data = self.preset_groups["Preset groups"][self.current_preset_group]
        self.update_preset_groups(block_signals=True)

        for index in range(self.tab_groups.count()):
            group_name = self.tab_groups.tabText(index)
            table = self.tab_groups.widget(index)
            table.model().model_data = self.current_data[group_name]
            table.model().layoutChanged.emit()

    def add_group(self, title, model_data):
        view_model = ViewModel(model_data)
        view_model.dataChanged.connect(self.data_changed)
        group_table = QTableView()
        group_table.setSelectionBehavior(group_table.SelectItems)
        group_table.setSelectionMode(group_table.SingleSelection)
        group_table.horizontalHeader().setSectionResizeMode(QHeaderView.ResizeToContents)
        group_table.horizontalHeader().setStretchLastSection(True)
        group_table.setModel(view_model)
        group_table.doubleClicked.connect(self.on_double_click)
        group_table.selectionModel().selectionChanged.connect(self.current_selection_changed)
        self.tab_groups.addTab(group_table, title)
        return view_model

    @pyqtSlot(QModelIndex, QModelIndex)
    def data_changed(self, from_index, to_index):
        table = self.tab_groups.currentWidget()
        new_value = table.model().model_data["rows"][from_index.row()][from_index.column() + 2]
        for index in range(self.tab_groups.count()):
            table = self.tab_groups.widget(index)
            table.model().model_data["rows"][from_index.row()][from_index.column() + 2] = new_value
            self.current_data[self.tab_groups.tabText(index)] = table.model().model_data

        self.update_source()

    @pyqtSlot(int)
    def current_tab_changed(self, tab_index):
        table = self.tab_groups.currentWidget()
        if table:
            group_name = str(self.tab_groups.tabText(self.tab_groups.currentIndex()))
            table.model().model_data = self.current_data[group_name]
            table.model().layoutChanged.emit()

            for button in self.manipulation_buttons:
                button.setDisabled(True)

            table.selectionModel().clearSelection()

    @pyqtSlot(QItemSelection, QItemSelection)
    def current_selection_changed(self, selected, deselected):
        if self.rows_selection_mode:
            for button in self.manipulation_buttons + [self.button_add]:
                button.setDisabled(True)
        else:
            self.button_add.setDisabled(False)
            for button in self.manipulation_buttons:
                button.setDisabled(True if len(selected) == 0 else False)

            if len(selected) > 0:
                if selected[0].topLeft().column() == 0:
                    self.button_set_value.setDisabled(True)

    @pyqtSlot(QModelIndex)
    def on_double_click(self, index):
        table = self.tab_groups.currentWidget()
        selected_index = table.selectedIndexes()[0]

        if self.checkbox_allow_double_click.isChecked() and selected_index.column() > 0:
            self.set_attribute_value()

    @pyqtSlot()
    def move_rows_up(self):
        table = self.tab_groups.currentWidget()
        selected_index = table.selectedIndexes()[0]
        if selected_index.row() == 0:
            return
        for index in range(self.tab_groups.count()):
            table = self.tab_groups.widget(index)
            row = table.model().model_data["rows"].pop(selected_index.row())
            table.model().model_data["rows"].insert(selected_index.row() - 1, row)
            table.model().layoutChanged.emit()
            table.setCurrentIndex(table.model().createIndex(selected_index.row() - 1, selected_index.column()))
            self.current_data[self.tab_groups.tabText(index)] = table.model().model_data

        self.update_source()

    @pyqtSlot()
    def move_rows_down(self):
        table = self.tab_groups.currentWidget()
        selected_index = table.selectedIndexes()[0]
        if selected_index.row() == table.model().rowCount() - 1:
            return
        for index in range(self.tab_groups.count()):
            table = self.tab_groups.widget(index)
            row = table.model().model_data["rows"].pop(selected_index.row())
            table.model().model_data["rows"].insert(selected_index.row() + 1, row)
            table.model().layoutChanged.emit()
            table.setCurrentIndex(table.model().createIndex(selected_index.row() + 1, selected_index.column()))
            self.current_data[self.tab_groups.tabText(index)] = table.model().model_data

        self.update_source()

    @pyqtSlot()
    def move_rows_top(self):
        table = self.tab_groups.currentWidget()
        selected_index = table.selectedIndexes()[0]
        if selected_index.row() == 0:
            return
        for index in range(self.tab_groups.count()):
            table = self.tab_groups.widget(index)
            row = table.model().model_data["rows"].pop(selected_index.row())
            table.model().model_data["rows"].insert(0, row)
            table.model().layoutChanged.emit()
            table.setCurrentIndex(table.model().createIndex(0, selected_index.column()))
            self.current_data[self.tab_groups.tabText(index)] = table.model().model_data

        self.update_source()

    @pyqtSlot()
    def move_rows_bottom(self):
        table = self.tab_groups.currentWidget()
        selected_index = table.selectedIndexes()[0]
        if selected_index.row() == table.model().rowCount():
            return
        for index in range(self.tab_groups.count()):
            table = self.tab_groups.widget(index)
            row = table.model().model_data["rows"].pop(selected_index.row())
            table.model().model_data["rows"].insert(table.model().rowCount(), row)
            table.model().layoutChanged.emit()
            table.setCurrentIndex(table.model().createIndex(table.model().rowCount() - 1, selected_index.column()))
            self.current_data[self.tab_groups.tabText(index)] = table.model().model_data

        self.update_source()

    @pyqtSlot()
    def remove_rows(self):
        answer = show_question_dialog("Do You really want to remove selected row ?", "Remove row", self)
        if answer:
            table = self.tab_groups.currentWidget()
            selected_index = table.selectedIndexes()[0]
            for index in range(self.tab_groups.count()):
                table = self.tab_groups.widget(index)
                table.model().model_data["rows"].pop(selected_index.row())
                table.model().layoutChanged.emit()
                self.current_data[self.tab_groups.tabText(index)] = table.model().model_data

            self.update_source()

    @pyqtSlot()
    def add_row(self):
        user_comment = show_prompt_dialog("Provide comment for new input:", "User comment", self)
        if user_comment:
            for index in range(self.tab_groups.count()):
                table = self.tab_groups.widget(index)
                saved_attributes = table.model().model_data["columns"]
                row = [self.default_row_color, time.strftime("%d-%m-%Y %H:%M:%S"), str(user_comment)]
                for device_id, attribute in saved_attributes[1:]:
                    device = self.session.get_device(device_id)
                    try:
                        if device.in_statuses(READY_DEVICE_STATUSES) and device.get_attribute(attribute) is not None:
                            value = device.get_attribute(attribute)
                            unit = device.get_attribute(attribute[:-1] + [UNIT])
                            user_unit = AttributeDeviceDisplayWidget.user_attribute_units.get(
                                (device_id, tuple(attribute[:-1])), None)
                            if user_unit is not None:
                                row.append("{:~}".format(units.Quantity(value, unit).to(user_unit)))
                            else:
                                row.append("{} {}".format(value, unit))
                        else:
                            row.append(u"None")
                    except DeviceUnknownError:
                        row.append(u"None")
                table.model().model_data["rows"].append(row)
                table.model().layoutChanged.emit()
                self.current_data[self.tab_groups.tabText(index)] = table.model().model_data

            self.update_source()

    @pyqtSlot()
    def save(self):
        extension_filter = "SaveAttributes [*.att] (*.att)"
        file_path, _ = QFileDialog.getSaveFileName(self, 'Set save file', "./%s.att" % "attributeValues",
                                                   filter=extension_filter, initialFilter=extension_filter)
        if file_path != "":
            with open(file_path, "w") as fp:
                yaml.dump(self.current_data, fp)

    @pyqtSlot()
    def load(self):
        extension_filter = "SaveAttributes [*.att] (*.att)"
        file_path, _ = QFileDialog.getOpenFileName(self, 'Choose file to load', ".", filter=extension_filter,
                                                   initialFilter=extension_filter)
        if file_path != "":
            with open(file_path, "r") as fp:
                self.file_path = file_path

                if self.device is not None:
                    self.device.detach_observer(self)
                    self.device.detach_attribute_callback(ATTR_CONTENT, self.content_update)
                    self.device = None
                    self.checkbox_synced_with_server.setChecked(False)

                self.preset_groups = yaml.load(fp, Loader=yaml.Loader)
                self.current_data = self.preset_groups["Preset groups"][self.current_preset_group]
                self.load_data(self.current_data)

    @pyqtSlot()
    def reset(self):
        answer = show_question_dialog("Do You really want to remove all rows ?", "Reset all groups", self)
        if answer:
            table = self.tab_groups.currentWidget()
            table.model().model_data["rows"] = []
            table.model().layoutChanged.emit()

            for button in self.manipulation_buttons:
                button.setDisabled(True)

            self.update_source()

    @pyqtSlot()
    def reset_all(self):
        answer = show_question_dialog("Do You really want to remove all rows ?", "Reset group", self)
        if answer:
            for index in range(self.tab_groups.count()):
                table = self.tab_groups.widget(index)
                table.model().model_data["rows"] = []
                table.model().layoutChanged.emit()

            for button in self.manipulation_buttons:
                button.setDisabled(True)

            self.update_source()

    @pyqtSlot()
    def select_color(self):
        color = QColorDialog().getColor()
        if color is not None and color.name() != "#000000":
            table = self.tab_groups.currentWidget()
            selected_row = table.selectedIndexes()[0].row()

            for index in range(self.tab_groups.count()):
                table = self.tab_groups.widget(index)
                table.model().model_data["rows"][selected_row][0] = str(color.name())
                self.current_data[self.tab_groups.tabText(index)] = table.model().model_data

            self.update_source()

    @pyqtSlot()
    def set_attribute_value(self):
        table = self.tab_groups.currentWidget()
        selected_index = table.selectedIndexes()[0]

        attribute_value = table.model().model_data["rows"][selected_index.row()][selected_index.column() + 2]
        device_id, attribute = table.model().model_data["columns"][selected_index.column()]

        if attribute_value == "None":
            error_message = u"Cannot set attribute {} to value {}".format(attribute, attribute_value)
            show_error_message(error_message, self)
            self.logger.info(error_message)
            return
        else:
            self.logger.info(u"Set attribute {} to value {}".format(attribute, attribute_value))

        try:
            device = self.session.get_device(device_id)
            value_in_device_units = device_units(device, attribute[:-1], attribute_value)
            device.set_attribute(attribute, value_in_device_units.m, callback=self.sigErrorCheck.emit)
        except (DeviceError, DimensionalityError) as e:
            self.logger.exception(u"Set attribute error")
            show_error_message(e, self)

    @pyqtSlot()
    def add_preset_group(self):
        group_name = show_prompt_dialog("Set name of a new group", "New save group")
        if group_name and group_name != "":
            new_group = {}
            for key, value in self.preset_groups["Preset groups"][self.current_preset_group].items():
                new_group[key] = {
                    "columns": value["columns"],
                    "rows": []
                }
            self.preset_groups["Preset groups"][group_name] = new_group
            self.current_preset_group = group_name
            self.update_source()

    @pyqtSlot()
    def remove_preset_group(self):
        if show_question_dialog("Do You really want to remove current group?"):
            self.preset_groups["Preset groups"].pop(self.current_preset_group, None)
            self.combo_preset.removeItem(self.combo_preset.currentIndex())
            self.update_source()

    @pyqtSlot()
    def add_preset_group_from_selected(self):
        self.rows_selection_mode = True
        for button in self.manipulation_buttons + [self.button_add]:
            button.setDisabled(True)
        for button in self.preset_group_buttons:
            button.hide()
        for button in self.preset_rows_selection_buttons:
            button.show()

        for index in range(self.tab_groups.count()):
            table = self.tab_groups.widget(index)
            table.setSelectionBehavior(table.SelectRows)
            table.setSelectionMode(table.MultiSelection)

    def update_preset_groups(self, block_signals):
        if block_signals:
            self.combo_preset.blockSignals(True)
        preset_groups = list(self.preset_groups["Preset groups"].keys())
        self.combo_preset.clear()
        self.combo_preset.insertItems(0, preset_groups)
        self.combo_preset.setCurrentIndex(preset_groups.index(self.current_preset_group))
        if block_signals:
            self.combo_preset.blockSignals(False)

    @pyqtSlot("QString")
    def preset_group_changed(self, preset_goup):
        self.current_preset_group = preset_goup
        if self.current_preset_group == "Default":
            self.button_remove_preset_group.setDisabled(True)
        else:
            self.button_remove_preset_group.setDisabled(False)
        self.load_data(self.preset_groups)

    @pyqtSlot()
    def confirm_rows_selection(self):
        group_name = show_prompt_dialog("Set name of a new group", "New save group")
        if group_name and group_name != "":
            table = self.tab_groups.currentWidget()
            new_group = {}
            for key, value in self.preset_groups["Preset groups"][self.current_preset_group].items():
                rows = []
                for index in table.selectionModel().selectedRows():
                    rows.append(value["rows"][index.row()])
                new_group[key] = {
                    "columns": value["columns"],
                    "rows": rows
                }
            self.preset_groups["Preset groups"][group_name] = new_group
            self.current_preset_group = group_name
            self.update_source()

        self.cancel_rows_selection()

    @pyqtSlot()
    def cancel_rows_selection(self):
        self.rows_selection_mode = False
        for button in self.preset_group_buttons:
            button.show()
        for button in self.preset_rows_selection_buttons:
            button.hide()

        for index in range(self.tab_groups.count()):
            table = self.tab_groups.widget(index)
            table.setSelectionBehavior(table.SelectItems)
            table.setSelectionMode(table.SingleSelection)
            table.clearSelection()



import inspect
from collections import OrderedDict

import oyaml as yaml


class YamlDumper(yaml.Dumper):

    def represent_data(self, data):
        """
        Custom Yaml representer.
        Every class inheriting from YamlSerializable is serialized using this representer.
        Use traditional serialization on other classes.
        :param data: mixed
        :return: Yaml Node
        """
        device_py_path = "{}.{}".format(data.__class__.__module__, data.__class__.__name__)

        if isinstance(data, YamlSerializable):
            if self.ignore_aliases(data):
                self.alias_key = None
            else:
                self.alias_key = id(data)
            if self.alias_key is not None:
                if self.alias_key in self.represented_objects:
                    node = self.represented_objects[self.alias_key]
                    # if node is None:
                    #    raise RepresenterError("recursive objects are not allowed: %r" % data)
                    return node
                # self.represented_objects[alias_key] = None
                self.object_keeper.append(data)

            output_node = self.represent_mapping(u"!Device:{}".format(device_py_path), data.yaml_mapping(),
                                                 flow_style=False)
            # self.anchor_node(output_node)
            return output_node
        else:
            try:
                return yaml.Dumper.represent_data(self, data)
            except TypeError:
                return yaml.Dumper.represent_none(self, "")


class YamlSerializable(object):

    def yaml_mapping(self):
        """
        Returns mapping of all arguments for __init__ function.
        If You want to return custom Yaml mapping re-implement this method.
        :return: OrderedDict
        """
        arg_spec = inspect.signature(self.__init__).parameters.keys()
        mapping = OrderedDict()
        for arg in arg_spec:
            try:
                argument_value = self.__getattribute__(arg)
                if argument_value is not None:
                    mapping[arg] = argument_value
            except (KeyError, AttributeError):
                """
                Key does not exists, skip in that case
                If You want to implement missing argument, re-implement yaml_mapping in
                particular device class.
                """
                pass
        return mapping



import sys
from collections import deque
from threading import Lock, Thread
from time import sleep

from PyQt5.QtCore import pyqtSignal, QObject, pyqtSlot
from PyQt5.QtWidgets import QWidget

from kamzik3.gui.templates.consoleTemplate import Ui_Form
from kamzik3.snippets.snippetsYaml import YamlSerializable


class OutputStream(QObject):
    sig_write = pyqtSignal("QString")

    def write(self, message):
        self.sig_write.emit(message)


class ConsoleWidget(Ui_Form, QWidget, YamlSerializable):
    sig_stdout_write = pyqtSignal("QString")
    sig_stderr_write = pyqtSignal("QString")
    sig_flush_output = pyqtSignal()
    info_color = "blue"
    error_color = "red"
    warning_color = "orange"
    debug_color = "green"
    output_paused = False

    def __init__(self, title=None, config=None, parent=None):
        QWidget.__init__(self, parent)
        self.config = config
        if self.config is None:
            self.config = {}
        self.output_lock = Lock()
        self.output_buffer = deque(maxlen=500)
        self.message_counter = 0
        self.error_counter = 0
        self.title = title
        self.previous_stdout = None
        self.previous_stderr = None
        if self.config.get("clone", True):
            self.previous_stdout = sys.stdout
            self.previous_stderr = sys.stderr
        sys.stdout = OutputStream()
        sys.stderr = OutputStream()

        self.sig_flush_output.connect(self.flush_output_buffer)
        sys.stdout.sig_write.connect(self.slot_write_stdout)
        sys.stderr.sig_write.connect(self.slot_write_stderr)
        self.setupUi(self)

    def yaml_mapping(self):
        mapping = super().yaml_mapping()
        del mapping["parent"]
        return mapping

    @pyqtSlot("QString")
    def slot_write_stdout(self, text):
        if not self.output_paused:
            if text != "\n":
                text = "> " + text
                self.output_buffer.appendleft(self.format_html(text.strip()))
                self.message_counter += 1

        if self.previous_stdout is not None:
            self.previous_stdout.write(text)

    @pyqtSlot("QString")
    def slot_write_stderr(self, text):
        if not self.output_paused:
            if text != "\n":
                self.output_buffer.appendleft(self.format_html(text.strip()))
                self.message_counter += 1

        if self.previous_stderr is not None:
            self.previous_stderr.write(text)

    @pyqtSlot()
    def flush_output_buffer(self):
        if self.message_counter != self.input_messages_count.value():
            output = "".join(self.output_buffer)
            self.input_console_text.setHtml(output)
            self.input_messages_count.setValue(self.message_counter)
            self.input_errors_count.setValue(self.error_counter)

    @pyqtSlot()
    def scroll_down(self):
        vs = self.input_console_text.verticalScrollBar()
        vs.setValue(vs.maximum())

    @pyqtSlot()
    def scroll_up(self):
        vs = self.input_console_text.verticalScrollBar()
        vs.setValue(vs.minimum())

    @pyqtSlot()
    def clear_all(self):
        self.message_counter = 0
        self.error_counter = 0
        self.input_console_text.clear()
        self.input_messages_count.setValue(0)
        self.input_errors_count.setValue(0)

    @pyqtSlot()
    def pause_output(self):
        self.output_paused = True
        self.button_resume_output.setEnabled(True)
        self.button_pause_output.setEnabled(False)

    @pyqtSlot()
    def resume_output(self):
        self.output_paused = False
        self.button_resume_output.setEnabled(False)
        self.button_pause_output.setEnabled(True)

    def _move_cursor_to_end(self):
        tc = self.input_console_text.textCursor()
        tc.setPosition(self.input_console_text.document().characterCount() - 1)
        self.input_console_text.setTextCursor(tc)

    def _move_cursor_to_beginning(self):
        tc = self.input_console_text.textCursor()
        tc.setPosition(0)
        self.input_console_text.setTextCursor(tc)

    def format_html(self, text):
        if ", INFO," in text:
            text = "<span style='color:{}'>{}</span><br/>".format("blue", text)
        elif ", DEBUG," in text:
            text = "<span style='color:{}'>{}</span><br/>".format("green", text)
        elif ", WARNING," in text:
            text = "<span style='color:{}'>{}</span><br/>".format("orange", text)
        elif ", ERROR," in text:
            self.error_counter += 1
            text = "<span style='color:{}'>{}</span><br/>".format("red", text)
        else:
            text = "<span>{}</span>".format(text)
        return text

    def showEvent(self, QShowEvent):

        def flush_output_thread():
            while self.isVisible():
                self.sig_flush_output.emit()
                sleep(2)

        Thread(target=flush_output_thread).start()



from kamzik3.gui.attributeWidgets.attributeFloatWidget import AttributeFloatWidget


class AttributeUnsignedIntWidget(AttributeFloatWidget):
    attribute_type_cast = int

    def _set_input_widget(self):
        super()._set_input_widget()
        self.input_widget.setDecimals(0)
        value_min = self.attribute.minimum()
        if value_min >= 0:
            self.input_widget.setMinimum(self.attribute_type_cast(value_min.m))
        else:
            self.input_widget.setMinimum(0)



import time
from collections import deque

from PyQt5.QtCore import pyqtSignal, pyqtSlot
from pint import DimensionalityError

from kamzik3 import units, DeviceError
from kamzik3.constants import *
from kamzik3.devices.attribute import Attribute
from kamzik3.devices.device import Device
from kamzik3.gui.attributeDisplayWidget import AttributeDisplayWidget
from kamzik3.gui.deviceWidget import DeviceWidget
from kamzik3.snippets.snippetsWidgets import show_error_message


class AttributeDeviceDisplayWidget(DeviceWidget, AttributeDisplayWidget):
    sigAttributeChanged = pyqtSignal("PyQt_PyObject", "PyQt_PyObject")
    attribute = None
    # Global history for every device display attribute
    setpoints_history = deque(maxlen=100)
    # Global dictionary for last device, attribute unit
    user_attribute_units = {}

    def __init__(self, device, attribute, config=None, parent=None):
        self.initial_attribute = Attribute.list_attribute(attribute)
        self.config = config
        if self.config is None:
            self.config = {}
        self.title = self.initial_attribute[-1]

        if isinstance(device, Device):
            self.attribute_device_key = (device.device_id, tuple(self.initial_attribute))
        else:
            self.attribute_device_key = (device, tuple(self.initial_attribute))

        DeviceWidget.__init__(self, device, config=config, parent=parent)

    def init_signals(self):
        super().init_signals()
        self.sigAttributeChanged.connect(self.slot_attribute_changed)

    def setupUi(self, parent=None):
        return

    def on_attribute_changed(self, key, value):
        self.sigAttributeChanged.emit(key, value)

    def get_current_user_unit(self):
        return AttributeDeviceDisplayWidget.user_attribute_units.get(self.attribute_device_key, None)

    @pyqtSlot("QString")
    def set_unit(self, unit=None, reset=False):
        self.attribute_widget.set_unit(unit, reset)

    @pyqtSlot()
    def slot_handle_configuration(self):
        DeviceWidget.slot_handle_configuration(self)

        self.attribute = self.device.get_attribute(self.initial_attribute)
        AttributeDisplayWidget.setupUi(self)

        current_user_unit = self.get_current_user_unit()
        if current_user_unit is not None:
            self.unit_widget.set_unit(current_user_unit)

        self.unit_widget.sig_unit_changed.connect(self.set_unit)
        self.unit_widget.sig_unit_changed.connect(self.on_user_unit_changed)
        self.input_widget.reset_on_blur = True
        self.slot_observing_start()

    @pyqtSlot("QString")
    def on_user_unit_changed(self, unit):
        AttributeDeviceDisplayWidget.user_attribute_units[self.attribute_device_key] = unit

    @pyqtSlot("bool")
    def slot_set_enabled(self, value):
        self.input_widget.setEnabled(value)
        self.setEnabled(value)

    @pyqtSlot()
    def slot_observing_start(self):
        if self.device is not None:
            self.device.attach_attribute_callback(self.initial_attribute, self.on_attribute_changed,
                                                  self.config.get("max_update_rate"))
            self.device.attach_attribute_callback(ATTR_ENABLED, self.on_enabled_changed)
            self.setEnabled(self.device.get_value(ATTR_ENABLED))
            self.update_value()

    @pyqtSlot()
    def slot_observing_stop(self):
        if self.device is not None:
            self.device.detach_attribute_callback(self.initial_attribute, self.on_attribute_changed)
            self.device.detach_attribute_callback(ATTR_ENABLED, self.on_enabled_changed)

    def update_value(self):
        if self.attribute is not None:
            self.set_value(self.attribute.value())

    # def hideEvent(self, hide_event):
    #     self.slot_observing_stop()
    #     DeviceWidget.hideEvent(self, hide_event)
    #
    def showEvent(self, show_event):
        self.update_value()
        # self.slot_observing_start()
        # DeviceWidget.showEvent(self, show_event)

    def _add_in_setpoint_history(self, value):
        """
        Add setpoint into setpoint setpoints_history list.
        Each input is in format (epoch time, device ID, attribute name, previous value, current setpoint)
        :param value:
        :return:
        """
        previous_value = self.attribute.value()
        if previous_value is not None:
            previous_value = previous_value.to(self.get_unit())
        self.setpoints_history.append(
            (time.time(), self.device_id, self.initial_attribute, previous_value, value))

    def on_value_set(self, value):
        if isinstance(value, units.Quantity):
            self._add_in_setpoint_history(value)
            value = value.to(self.attribute[UNIT])
            self.logger.debug(u"Set {} to {}".format(self.initial_attribute, value))
            value = value.m
        elif self.attribute.numerical and self.get_unit() is not None:
            try:
                value = units.Quantity(value, self.get_unit())
                self._add_in_setpoint_history(value)
                self.logger.debug(u"Set {} to {}".format(self.initial_attribute, value))
                value = value.to(self.attribute[UNIT]).m
            except AttributeError:
                pass
        elif self.attribute[TYPE] == TYPE_LIST:
            value = self.get_widget_value()

        try:
            self.device.set_attribute(self.initial_attribute + [VALUE],
                                      self.attribute_widget.attribute_type_cast(value),
                                      callback=self.sigErrorCheck.emit)
        except DeviceError as e:
            # Set value back to what it was before
            self.set_value(self.attribute[VALUE])
            show_error_message(e, parent=self)

        self.input_widget.setFocus()
        self.input_widget.clearFocus()
        self.parent().clearFocus()

    @pyqtSlot("PyQt_PyObject")
    def slot_set_status(self, value):
        DeviceWidget.slot_set_status(self, value)
        if not self.isVisible():
            return
        elif value in READY_DEVICE_STATUSES:
            self.setEnabled(self.device.get_value(ATTR_ENABLED))
        elif value in (STATUS_DISCONNECTED, STATUS_DISCONNECTING):
            self.setDisabled(True)

    @pyqtSlot("PyQt_PyObject", "PyQt_PyObject")
    def slot_attribute_changed(self, key, value):
        if self.attribute is None:
            return
        if self.isVisible() and key == VALUE:
            if self.input_widget is not None and self.input_widget.hasFocus() or value is None:
                return
            if self.attribute.numerical:
                try:
                    self.set_value(units.Quantity(value, self.attribute[UNIT]))
                except AttributeError:
                    self.set_value(units.Quantity(value))
                except DimensionalityError:
                    self.set_value(units.Quantity(units.Quantity(value).m, self.get_unit()))
            else:
                self.set_value(value)
        elif key == UNIT:
            self.attribute_widget.unit = value
            self.unit_widget.set_unit(value)
            self.set_unit(value, True)
            self.set_minimum(units.Quantity(self.attribute[MIN], value))
            self.set_maximum(units.Quantity(self.attribute[MAX], value))
            self.slot_attribute_changed(VALUE, self.attribute[VALUE])
        elif key == MIN:
            self.set_minimum(units.Quantity(value, self.attribute[UNIT]))
        elif key == MAX:
            self.set_maximum(units.Quantity(value, self.attribute[UNIT]))
        elif key == FACTOR:
            self.slot_attribute_changed(VALUE, self.attribute[VALUE])

    def close(self):
        self.sigAttributeChanged.disconnect()
        try:
            self.device.detach_attribute_callback(self.initial_attribute, self.on_attribute_changed)
            self.device.detach_attribute_callback(ATTR_ENABLED, self.on_enabled_changed)
        except:
            pass
        self.initial_attribute = None
        self.config = None
        DeviceWidget.close(self)
        AttributeDisplayWidget.close(self)



import logging

from PyQt5.QtCore import pyqtSlot, pyqtSignal
from pyqtgraph import GraphicsLayoutWidget

import kamzik3
from kamzik3 import units
from kamzik3.constants import *
from kamzik3.gui.attributePlot import AttributePlot
from kamzik3.gui.deviceWidget import DeviceWidget
from kamzik3.gui.templates.macroProgressTemplate import Ui_Form
from kamzik3.snippets.snippetsWidgets import show_error_message


class MacroProgressWidget(Ui_Form, DeviceWidget):
    sig_update_progress = pyqtSignal("float")
    sig_update_plot = pyqtSignal(object)
    sig_update_macro_status = pyqtSignal("QString")

    def __init__(self, device, template, model_image=None, config=None, parent=None):
        self.template = template
        self.plot_widgets = []
        self.plot_area = GraphicsLayoutWidget()
        self.logger = logging.getLogger("Gui.MacroProgressWidget")
        super().__init__(device, model_image, config, parent)
        self.plot_widget.layout().addWidget(self.plot_area)

    def _init_macro_plots(self):
        for index, value in enumerate(self.device.get_chain_link_output()):
            if value is not None:
                (device_id, attribute_name, points) = value
                device = kamzik3.session.get_device(device_id)
                attribute = device.get_attribute(attribute_name)
                left_label = "{} - {}".format(device_id, attribute_name)
                plot = self.plot_area.addPlot(index, 0)
                attribute_plot = AttributePlot(attribute, buffer_size=points, left_label=left_label, plot=plot)
                self.plot_widgets.append(attribute_plot)
            else:
                self.plot_widgets.append(None)

    @pyqtSlot()
    def slot_handle_configuration(self):
        """
        Handle widget configuration.
        If overloading make sure to set configured flag.
        :return:
        """
        self._init_macro_plots()
        self.device.attach_attribute_callback(ATTR_PROGRESS, self.progress_update)
        self.device.attach_attribute_callback(ATTR_CHAIN_LINK_OUTPUT, self.plot_update)
        self.sig_update_progress.connect(self.set_macro_progress)
        self.sig_update_plot.connect(self.set_plot_update)
        DeviceWidget.slot_handle_configuration(self)

    @pyqtSlot("float")
    def set_macro_progress(self, progress):
        self.progress_bar.setFormat("{0:.2f}%".format(progress))
        self.progress_bar.setValue(progress)
        self.label_macro_step_progress.setText(
            "  {} / {}  ".format(int(self.device[ATTR_POINT_INDEX][VALUE]),
                                 int(self.device[ATTR_POINTS_COUNT][VALUE])))
        self.label_running_time.setText("  {}  ".format(self.device[ATTR_RUNNING_TIME][VALUE]))
        self.label_remaining_time.setText("  {}  ".format(self.device[ATTR_REMAINING_TIME][VALUE]))
        self.label_estimated_finish.setText("  {}  ".format(self.device[ATTR_ESTIMATED_END][VALUE]))

    def set_plot_update(self, link_data):
        macro_point, chain_link, current_value = link_data
        if self.device[ATTR_POINT_INDEX][VALUE] > 0:
            plot_widget = self.plot_widgets[chain_link]
            if plot_widget is not None:
                current_value = units.Quantity(current_value)
                plot_widget.update(macro_point, current_value)

    @pyqtSlot("PyQt_PyObject")
    def slot_set_status(self, status):
        """
        This slot is always called when device status has changed.
        Recall thic method when overloading.
        :param status:
        :return:
        """
        if not self.configured and status in READY_DEVICE_STATUSES:
            self.slot_handle_configuration()
        if status == STATUS_BUSY:
            self.button_stop.setDisabled(False)
            self.button_start.setDisabled(True)
            for plot in self.plot_widgets:
                if plot is not None:
                    plot.reset()
        elif status in (STATUS_IDLE, STATUS_ERROR):
            self.button_stop.setDisabled(True)
            self.button_start.setDisabled(False)
        elif status not in READY_DEVICE_STATUSES:
            self.button_stop.setDisabled(True)
            self.button_start.setDisabled(True)

        self.label_status.setText(status)
        self.label_status.setStyleSheet(
            "QLabel {{background:{}}}".format(DeviceWidget.status_colors.get(status, "blue")))

    def progress_update(self, key, value):
        if key == VALUE:
            self.sig_update_progress.emit(value)

    def status_update(self, key, value):
        if key == VALUE:
            self.sig_update_macro_status.emit(value)

    def plot_update(self, key, value):
        if key == VALUE:
            self.sig_update_plot.emit(value)

    @pyqtSlot()
    def start(self):
        self.device.start()

    @pyqtSlot()
    def stop(self):
        self.device.stop()

    def close(self):
        self.device.detach_attribute_callback(ATTR_PROGRESS, self.progress_update)
        for plot in self.plot_widgets:
            if plot is None:
                continue
            plot.close()
            del plot
        del self.template
        del self.plot_widgets
        del self.plot_area
        macro = self.device
        DeviceWidget.close(self)
        macro.remove()



from kamzik3.gui.attributeWidgets.attributeFloatWidget import AttributeFloatWidget


class AttributeIntWidget(AttributeFloatWidget):
    attribute_type_cast = int

    def _set_input_widget(self):
        super()._set_input_widget()
        self.input_widget.setDecimals(0)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'modelTableTemplate.ui'
#
# Created by: PyQt5 UI code generator 5.13.2
#
# WARNING! All changes made in this file will be lost!


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(264, 326)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setObjectName("gridLayout")
        self.input_table_view = QtWidgets.QTableView(Form)
        self.input_table_view.setAlternatingRowColors(True)
        self.input_table_view.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.input_table_view.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectItems)
        self.input_table_view.setGridStyle(QtCore.Qt.SolidLine)
        self.input_table_view.setObjectName("input_table_view")
        self.input_table_view.horizontalHeader().setCascadingSectionResizes(True)
        self.input_table_view.horizontalHeader().setSortIndicatorShown(False)
        self.input_table_view.horizontalHeader().setStretchLastSection(True)
        self.input_table_view.verticalHeader().setCascadingSectionResizes(False)
        self.gridLayout.addWidget(self.input_table_view, 0, 0, 1, 1)
        self.layout_controls = QtWidgets.QVBoxLayout()
        self.layout_controls.setSizeConstraint(QtWidgets.QLayout.SetFixedSize)
        self.layout_controls.setObjectName("layout_controls")
        self.button_copy_item = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_copy_item.sizePolicy().hasHeightForWidth())
        self.button_copy_item.setSizePolicy(sizePolicy)
        self.button_copy_item.setMinimumSize(QtCore.QSize(32, 32))
        self.button_copy_item.setMaximumSize(QtCore.QSize(32, 32))
        self.button_copy_item.setText("")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/icons/copy.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_copy_item.setIcon(icon)
        self.button_copy_item.setIconSize(QtCore.QSize(25, 25))
        self.button_copy_item.setCheckable(False)
        self.button_copy_item.setObjectName("button_copy_item")
        self.layout_controls.addWidget(self.button_copy_item)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.layout_controls.addItem(spacerItem)
        self.button_move_item_top = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_move_item_top.sizePolicy().hasHeightForWidth())
        self.button_move_item_top.setSizePolicy(sizePolicy)
        self.button_move_item_top.setMinimumSize(QtCore.QSize(32, 32))
        self.button_move_item_top.setMaximumSize(QtCore.QSize(32, 32))
        self.button_move_item_top.setText("")
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/icons/icons/top.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_move_item_top.setIcon(icon1)
        self.button_move_item_top.setIconSize(QtCore.QSize(25, 25))
        self.button_move_item_top.setObjectName("button_move_item_top")
        self.layout_controls.addWidget(self.button_move_item_top)
        self.button_move_item_up = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_move_item_up.sizePolicy().hasHeightForWidth())
        self.button_move_item_up.setSizePolicy(sizePolicy)
        self.button_move_item_up.setMinimumSize(QtCore.QSize(32, 32))
        self.button_move_item_up.setMaximumSize(QtCore.QSize(32, 32))
        self.button_move_item_up.setText("")
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/icons/icons/plus.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_move_item_up.setIcon(icon2)
        self.button_move_item_up.setIconSize(QtCore.QSize(25, 25))
        self.button_move_item_up.setCheckable(False)
        self.button_move_item_up.setChecked(False)
        self.button_move_item_up.setObjectName("button_move_item_up")
        self.layout_controls.addWidget(self.button_move_item_up)
        spacerItem1 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.layout_controls.addItem(spacerItem1)
        self.button_add_item = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_add_item.sizePolicy().hasHeightForWidth())
        self.button_add_item.setSizePolicy(sizePolicy)
        self.button_add_item.setMinimumSize(QtCore.QSize(32, 32))
        self.button_add_item.setMaximumSize(QtCore.QSize(32, 32))
        font = QtGui.QFont()
        font.setPointSize(14)
        self.button_add_item.setFont(font)
        self.button_add_item.setStyleSheet("QPushButton {color:green}")
        self.button_add_item.setObjectName("button_add_item")
        self.layout_controls.addWidget(self.button_add_item)
        self.button_remove_item = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_remove_item.sizePolicy().hasHeightForWidth())
        self.button_remove_item.setSizePolicy(sizePolicy)
        self.button_remove_item.setMinimumSize(QtCore.QSize(32, 32))
        self.button_remove_item.setMaximumSize(QtCore.QSize(32, 32))
        font = QtGui.QFont()
        font.setPointSize(14)
        self.button_remove_item.setFont(font)
        self.button_remove_item.setStyleSheet("QPushButton {color:red}")
        self.button_remove_item.setObjectName("button_remove_item")
        self.layout_controls.addWidget(self.button_remove_item)
        spacerItem2 = QtWidgets.QSpacerItem(20, 20, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Fixed)
        self.layout_controls.addItem(spacerItem2)
        self.button_move_item_down = QtWidgets.QPushButton(Form)
        self.button_move_item_down.setEnabled(True)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_move_item_down.sizePolicy().hasHeightForWidth())
        self.button_move_item_down.setSizePolicy(sizePolicy)
        self.button_move_item_down.setMinimumSize(QtCore.QSize(32, 32))
        self.button_move_item_down.setMaximumSize(QtCore.QSize(32, 32))
        self.button_move_item_down.setText("")
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/icons/icons/minus.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_move_item_down.setIcon(icon3)
        self.button_move_item_down.setIconSize(QtCore.QSize(25, 25))
        self.button_move_item_down.setObjectName("button_move_item_down")
        self.layout_controls.addWidget(self.button_move_item_down)
        self.button_move_item_bottom = QtWidgets.QPushButton(Form)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.button_move_item_bottom.sizePolicy().hasHeightForWidth())
        self.button_move_item_bottom.setSizePolicy(sizePolicy)
        self.button_move_item_bottom.setMinimumSize(QtCore.QSize(32, 32))
        self.button_move_item_bottom.setMaximumSize(QtCore.QSize(32, 32))
        self.button_move_item_bottom.setText("")
        icon4 = QtGui.QIcon()
        icon4.addPixmap(QtGui.QPixmap(":/icons/icons/bottom.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.button_move_item_bottom.setIcon(icon4)
        self.button_move_item_bottom.setIconSize(QtCore.QSize(25, 25))
        self.button_move_item_bottom.setObjectName("button_move_item_bottom")
        self.layout_controls.addWidget(self.button_move_item_bottom)
        spacerItem3 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.layout_controls.addItem(spacerItem3)
        self.gridLayout.addLayout(self.layout_controls, 0, 1, 1, 1)

        self.retranslateUi(Form)
        self.button_add_item.clicked.connect(Form.slot_add_item)
        self.button_remove_item.clicked.connect(Form.slot_remove_item)
        self.button_move_item_top.clicked.connect(Form.slot_move_item_top)
        self.button_move_item_up.clicked.connect(Form.slot_move_item_up)
        self.button_move_item_down.clicked.connect(Form.slot_move_item_down)
        self.button_move_item_bottom.clicked.connect(Form.slot_move_item_bottom)
        self.button_copy_item.clicked.connect(Form.slot_copy_item)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.button_copy_item.setToolTip(_translate("Form", "Copy selected rows"))
        self.button_move_item_top.setToolTip(_translate("Form", "Move selected row on top"))
        self.button_move_item_up.setToolTip(_translate("Form", "Move selected row up"))
        self.button_add_item.setToolTip(_translate("Form", "Add row"))
        self.button_add_item.setText(_translate("Form", "+"))
        self.button_remove_item.setToolTip(_translate("Form", "Remove selected row"))
        self.button_remove_item.setText(_translate("Form", "-"))
        self.button_move_item_down.setToolTip(_translate("Form", "Move selected row down"))
        self.button_move_item_bottom.setToolTip(_translate("Form", "Move selected row to bottom"))
import resource_kamzik3_rc



from PyQt5.QtCore import pyqtSlot, pyqtSignal
from PyQt5.QtWidgets import QWidget

from kamzik3.constants import *
from kamzik3.devices.attribute import Attribute
from kamzik3.gui.attributeWidgets.attributeUnitWidget import AttributeUnitWidget


class AttributeWidget(QWidget):
    input_widget = None
    unit_widget = None
    unit = None
    font_color = u"black"
    background_color = u"white"
    attribute = None
    attribute_type_cast = str
    sig_unit_changed = pyqtSignal("QString", "QString")

    def __init__(self, attribute, config=None, parent=None):
        assert isinstance(attribute, Attribute)
        self.attribute = attribute
        self.config = config
        if self.config is None:
            self.config = {}
        QWidget.__init__(self, parent=parent)
        self.setupUi()

    def _set_unit_widget(self):
        self.unit_widget = AttributeUnitWidget(self.attribute)
        self.unit_widget.sig_unit_changed.connect(self.set_unit)

    def set_tooltip(self, value=None):
        tooltip_value = self.attribute[DESCRIPTION]
        if tooltip_value is not None:
            tooltip_value += "\n" + value
        else:
            tooltip_value = value

        if tooltip_value is not None:
            self.input_widget.setToolTip(tooltip_value)

    def update_value(self):
        self.set_value(self.attribute.value())

    @pyqtSlot("PyQt_PyObject")
    def _set_input_widget(self):
        raise NotImplementedError("Has to be implemented")

    def get_attribute_value(self):
        raise NotImplementedError("Has to be implemented")

    def get_widget_value(self):
        raise NotImplementedError("Has to be implemented")

    def set_value(self, value):
        raise NotImplementedError("Has to be implemented")

    @pyqtSlot("QString")
    def set_unit(self, unit=None, reset=False):
        return

    def get_unit(self):
        return self.unit

    def setDisabled(self, bool):
        self.input_widget.setDisabled(bool)
        super().setDisabled(bool)

    def setupUi(self, parent=None):
        self.unit = self.attribute[UNIT]
        self._set_input_widget()
        self._set_unit_widget()

        if self.attribute[VALUE] is not None:
            self.set_value(self.attribute.value())

        if self.config is not None and self.config.get("unit") is not None:
            self.unit_widget.set_unit(self.config.get("unit"))
        self.input_widget.__dict__["update_value"] = self.update_value

    def close(self):
        if self.input_widget is not None:
            self.input_widget.setParent(None)
            del self.input_widget
        if self.unit_widget is not None:
            self.unit_widget.close()

        self.attribute = None
        self.config = None
        return super().close()



import socket
import time
from _socket import timeout
from threading import Thread

import numpy as np
import serial

from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.devices.devicePort import DevicePort
from kamzik3.devices.deviceSocket import DeviceSocket
from kamzik3.snippets.snippetsDecorators import expose_method


class DeviceGalil(DeviceSocket):
    terminator = b":"
    analog_input_readout_command = "MG @AN[0],@AN[1],@AN[2],@AN[3],@AN[4],@AN[5],@AN[6],@AN[7];"
    analog_output_readout_command = "MG @AO[0],@AO[1],@AO[2],@AO[3],@AO[4],@AO[5],@AO[6],@AO[7];"
    analog_output_limits_command = "MG _DQ0,_DQ1,_DQ2,_DQ3,_DQ4,_DQ5,_DQ6,_DQ7;"
    analog_input_limits_command = "MG _AQ0,_AQ1,_AQ2,_AQ3,_AQ4,_AQ5,_AQ6,_AQ7;"
    digital_lO_readout_command = "MG _TI0,_TI1,_OP0,_OP1;"
    push_commands_max = 1
    response_timeout = 1000
    digital_mask = (1, 2, 4, 8, 16, 24, 48)

    def __init__(self, host, port, udp_host=None, udp_port=None, device_id=None, config=None):
        self.udp_host = udp_host
        self.udp_port = udp_port
        self.udp_socket = None
        self.DO_config = []
        self.DI_config = []
        self.AO_config = []
        self.AI_config = []
        self.analog_input_date_type = np.dtype(
            [("AI0", np.ushort), ("AI1", np.ushort), ("AI2", np.ushort), ("AI3", np.ushort), ("AI4", np.ushort),
             ("AI5", np.ushort), ("AI6", np.ushort), ("AI7", np.ushort)])
        self.analog_output_date_type = np.dtype(
            [("AO0", np.ushort), ("AO1", np.ushort), ("AO2", np.ushort), ("AO3", np.ushort), ("AO4", np.ushort),
             ("AO5", np.ushort), ("AO6", np.ushort), ("AO7", np.ushort)])
        self.preset_io_attributes()
        super().__init__(host, port, device_id, config)

    def _init_attributes(self):
        super(DeviceGalil, self)._init_attributes()

        self.create_attribute(ATTR_MODEL, default_value="", readonly=True)

        self["Analog input"] = {}
        self["Analog output"] = {}
        self["Digital input"] = {}
        self["Digital output"] = {}

        for i in range(8):
            self.create_attribute(self.AI_config[i], "Analog input", min_value=0., max_value=10., unit=u"V",
                                  default_type=np.float16, decimals=4, readonly=True)
            self.create_attribute(self.AO_config[i], "Analog output", min_value=0., max_value=10., unit=u"V",
                                  set_function=lambda value, output=i: self.set_analog_output(output, value),
                                  default_type=np.float16, decimals=4)
            self.create_attribute(self.DI_config[i], "Digital input", default_type=np.bool, readonly=True)
            self.create_attribute(self.DI_config[8 + i], "Digital input", default_type=np.bool, readonly=True)
            self.create_attribute(self.DO_config[i], "Digital output", default_type=np.bool,
                                  set_value_when_set_function=False,
                                  set_function=lambda value, output=i: self.set_digital_output(output, value))
            self.create_attribute(self.DO_config[8 + i], "Digital output", default_type=np.bool,
                                  set_value_when_set_function=False,
                                  set_function=lambda value, output=8 + i: self.set_digital_output(output, value))

    def preset_io_attributes(self):
        self.DO_config = [
            "DO[0]", "DO[1]", "DO[2]", "DO[3]", "DO[4]", "DO[5]", "DO[6]", "DO[7]", "DO[8]", "DO[9]", "DO[10]",
            "DO[11]", "DO[12]", "DO[13]", "DO[14]", "DO[15]",
        ]
        self.DI_config = [
            "DI[0]", "DI[1]", "DI[2]", "DI[3]", "DI[4]", "DI[5]", "DI[6]", "DI[7]", "DI[8]", "DI[9]", "DI[10]",
            "DI[11]", "DI[12]", "DI[13]", "DI[14]", "DI[15]",
        ]
        self.AO_config = [
            "AO[0]", "AO[1]", "AO[2]", "AO[3]", "AO[4]", "AO[5]", "AO[6]", "AO[7]",
        ]
        self.AI_config = [
            "AI[0]", "AI[1]", "AI[2]", "AI[3]", "AI[4]", "AI[5]", "AI[6]", "AI[7]",
        ]

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

        self.command("EO 0;")
        self.command("IHC=>-1;")  # Disconnect any UDP connection on C interface
        self.command("DR0;")  # Stop any data recording

        if self.udp_port is not None:
            self.command("IHC={} <{} >1;".format(self.udp_host.replace(".", ","),
                                                 self.udp_port))  # Set UDP connection to provided host and port
            self.command("CFC;")  # Set interface C as main output for any UDP packet
            self.command("DR200,2;")  # Start recording data at 200ms and send it to second interface (C)
            self.udp_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.udp_socket.bind((self.udp_host, self.udp_port))
            self.udp_socket.settimeout(self.response_timeout * 1e-3)
            Thread(target=self._udp_readout).start()

        self.command("\x12\x16;")  # Get galil model and firmware version
        # self.command("TZ;") # Get all inputs and outputs
        self.command(self.analog_output_limits_command)
        self.command(self.analog_input_limits_command)
        self.command(self.analog_input_readout_command)
        self.command(self.analog_output_readout_command)
        self.command(self.digital_lO_readout_command, callback=_finish_configuration)

    def _udp_readout(self):
        self.request_timestamp = time.time()
        while self.connected:
            try:
                data, _ = self.udp_socket.recvfrom(1024)  # buffer size is 1024 bytes
                self._handle_udp_readout(data)
                self.response_timestamp = time.time()
            except timeout:
                self.request_timestamp = time.time()
                # DevicePoller.check_device_alive(self)
            except OSError:
                # Socket doesn't exist
                pass
        try:
            self.udp_socket.close()
            self.udp_socket = None
        except AttributeError:
            pass  # udp socket was already closed

    def _handle_udp_readout(self, data):
        try:
            model_name, revision = self.get_value(ATTR_MODEL).split(" ", maxsplit=1)
            if model_name == u"RIO47100":
                for index, ao in enumerate(np.frombuffer(data[8:24], dtype=self.analog_output_date_type)[0]):
                    ao_attribute = self["Analog output"][self.AO_config[index]]
                    ao_min, ao_max = ao_attribute.remove_offset_factor(
                        ao_attribute[MIN]), ao_attribute.remove_offset_factor(ao_attribute[MAX])
                    dt_info = np.iinfo(self.analog_output_date_type[index])
                    self.set_raw_value(("Analog output", self.AO_config[index]),
                                       (((ao - dt_info.min) * (ao_max - ao_min)) / (
                                                   dt_info.max - dt_info.min)) + ao_min)

                for index, ai in enumerate(np.frombuffer(data[24:40], dtype=self.analog_input_date_type)[0]):
                    ai_attribute = self["Analog input"][self.AI_config[index]]
                    ai_min, ai_max = ai_attribute.remove_offset_factor(
                        ai_attribute[MIN]), ai_attribute.remove_offset_factor(ai_attribute[MAX])
                    dt_info = np.iinfo(self.analog_input_date_type[index])
                    self.set_raw_value(("Analog input", self.AI_config[index]),
                                       (((ai - dt_info.min) * (ai_max - ai_min)) / (
                                                   dt_info.max - dt_info.min)) + ai_min)

                digital_output = format(np.frombuffer(data[40:42], dtype=np.ushort)[0], "b")
                digital_input = format(np.frombuffer(data[42:44], dtype=np.ushort)[0], "b")
                digital_output = [bool(int(x)) for x in reversed(digital_output)] + (
                        [False] * (16 - len(digital_output)))
                digital_input = [bool(int(x)) for x in reversed(digital_input)] + ([False] * (16 - len(digital_input)))

                for i, do in enumerate(digital_output):
                    self.set_raw_value(["Digital output", self.DO_config[i]], do)
                for i, di in enumerate(digital_input):
                    self.set_raw_value(["Digital input", self.DI_config[i]], di)
        except ValueError:
            pass

    def start_polling(self):
        if self.udp_port is None:
            Device.start_polling(self)
            self.poll_command(self.analog_input_readout_command, 200)
            self.poll_command(self.analog_output_readout_command, 200)
            self.poll_command(self.digital_lO_readout_command, 200)

    def handle_readout(self, readout_buffer):
        command, output, callback, token = super(DeviceGalil, self).handle_readout(readout_buffer)
        # Strip white characters from output
        output = output.strip()
        if output and output[0] == "?":
            self.handle_command_error(command, output)
        elif command == u"TC 1;":
            readout_buffer = output
            self.handle_command_error(command, output)
        elif command == self.analog_input_readout_command:
            analog_inputs = map(float, output.split())
            for i, value in enumerate(analog_inputs):
                self.set_raw_value(("Analog input", self.AI_config[i]), value)
        elif command == self.analog_output_readout_command:
            analog_outputs = map(float, output.split())
            for i, value in enumerate(analog_outputs):
                self.set_raw_value(("Analog output", self.AO_config[i]), value)
        elif command == self.digital_lO_readout_command:
            di0, di1, do0, do1 = map(lambda v: int(float(v)), output.split())
            for bit in range(8):
                self.set_raw_value(("Digital input", self.DI_config[bit]), bool(di0 & (1 << bit)))
                self.set_raw_value(("Digital input", self.DI_config[8 + bit]), bool(di1 & (1 << bit)))
                self.set_raw_value(("Digital output", self.DO_config[bit]), bool(do0 & (1 << bit)))
                self.set_raw_value(("Digital output", self.DO_config[8 + bit]), bool(do1 & (1 << bit)))
        elif command == "\x12\x16;":
            self.set_raw_value(ATTR_MODEL, output)
        elif command == self.analog_output_limits_command:
            analog_outputs_limits = output.strip().split()
            # This is analog output channels limits mapping from Galil doc
            analog_limit_ranges = [(-10, 10), (0, 5), (0, 10), (-5, 5), (-10, 10)]
            new_date_type = []
            for index, limit_range_option in enumerate(analog_outputs_limits):
                # Get lower and uper limit for each analog output channel
                lower_limit, upper_limit = analog_limit_ranges[int(float(limit_range_option))]
                # Figure out data type
                dt = np.ushort
                # Append tuple representing analog output data type
                new_date_type.append((self.AO_config[index], dt))
                # Set both limits for analog channel
                self.set_attribute(("Analog output", self.AO_config[index], MIN), lower_limit)
                self.set_attribute(("Analog output", self.AO_config[index], MAX), upper_limit)
            # Set numpy data type, which is used to parse binary UDP readout
            self.analog_output_date_type = np.dtype(new_date_type)
        elif command == self.analog_input_limits_command:
            analog_inputs_limits = output.strip().split()
            # This is analog input channels limits mapping from Galil doc
            analog_limit_ranges = [(-10, 10), (-5, 5), (-10, 10), (0, 5), (0, 10)]
            new_date_type = []
            for index, limit_range_option in enumerate(analog_inputs_limits):
                # Get lower and uper limit for each analog output channel
                lower_limit, upper_limit = analog_limit_ranges[int(float(limit_range_option))]
                # Figure out data type
                dt = np.short if lower_limit < 0 else np.ushort
                # Append tuple representing analog output data type
                new_date_type.append((self.AO_config[index], dt))
                # Set both limits for analog channel
                self.set_attribute(("Analog input", self.AI_config[index], MIN), lower_limit)
                self.set_attribute(("Analog input", self.AI_config[index], MAX), upper_limit)
            # Set numpy data type, which is used to parse binary UDP readout
            self.analog_input_date_type = np.dtype(new_date_type)

        if callback is not None:
            self.handle_readout_callback(callback, command, readout_buffer)
        if token:
            self.device_server.push_message(self.device_id, [command, readout_buffer], token)

        return command, output, callback, token

    def set_analog_output(self, output, value, callback=None):
        """
        Set analog output to value 0-10V
        :param output: int
        :param value: float
        :param callback: callable
        """
        self.logger.debug(u"Set analogue AO{} to value {}".format(output, value))
        return self.command("AO{},{};".format(output, value), callback, with_token=True)

    def set_digital_output(self, output, value, callback=None):
        """
        Set digital output to value 0 or 1
        :param output: int
        :param value: bool
        :param callback: callable
        """
        self.logger.debug(u"Set digital output to value {}".format(value))
        if value == 0:
            return self.command("CB{};".format(output), callback, with_token=True)
        else:
            return self.command("SB{};".format(output), callback, with_token=True)

    @expose_method({"bitmask": "0-65535"})
    def set_digital_output_mask(self, bitmask):
        self.logger.debug(u"Set digital output mask to value {}".format(bitmask))
        bitmask = int(bitmask)
        self.command("OP{};".format(bitmask))

    def set_command_echo(self, value, callback=None):
        """
        Set printing command for serial port together with command output.
        Example:
            EO 0: MG 1; => '1.0000\r\n:'
            EO 1: MG 1; => 'MG1; 1.0000\r\n:'
        :param value:
        :param callback: callable
        """
        self.logger.debug(u"Set command echo to value {}".format(value))
        return self.command("EO {};".format(value), callback, with_token=True)

    def valid_command_format(self, command):
        if command is not None and command[-1] == ";":
            return True
        else:
            return False

    def collect_incoming_data(self, data):
        data = data.decode("utf-8")
        if data.find("?") != -1:
            (attribute, token, callback, returning), command_init_timestamp = self.commands_buffer.popleft()
            self.command(u"TC 1;", with_token=token, callback=callback)
            data = data[data.find("?") + 2:]

        self.buffer.append(data)

    def get_error_code(self, callback=None):
        self.command(u"TC 1;", callback=callback)

    def close_connection(self):
        super().close_connection()
        if self.udp_socket is not None:
            self.udp_socket.close()


class DeviceGalilPort(DevicePort, DeviceGalil):
    terminator = DeviceGalil.terminator

    def __init__(self, port, baud_rate=115200, parity=serial.PARITY_NONE, stop_bits=serial.STOPBITS_ONE,
                 byte_size=serial.EIGHTBITS, device_id=None, udp_host=None, udp_port=None, config=None):
        self.udp_host = udp_host
        self.udp_port = udp_port
        self.udp_socket = None
        self.DO_config = []
        self.DI_config = []
        self.AO_config = []
        self.AI_config = []
        self.analog_input_date_type = np.dtype(
            [("AI0", np.ushort), ("AI1", np.ushort), ("AI2", np.ushort), ("AI3", np.ushort), ("AI4", np.ushort),
             ("AI5", np.ushort), ("AI6", np.ushort), ("AI7", np.ushort)])
        self.analog_output_date_type = np.dtype(
            [("AO0", np.ushort), ("AO1", np.ushort), ("AO2", np.ushort), ("AO3", np.ushort), ("AO4", np.ushort),
             ("AO5", np.ushort), ("AO6", np.ushort), ("AO7", np.ushort)])
        self.preset_io_attributes()
        DevicePort.__init__(self, port, baud_rate, parity, stop_bits, byte_size, device_id, config)

    def handshake(self):
        self.logger.info(u"Handshake initiated")
        self.push(b"MG 1;")
        time.sleep(0.02)  # Give some time for devices to reply
        if self.serial_port.in_waiting != 0:
            if self.read_all().strip() in (u"1.0000\r\n:", u"MG 1; 1.0000\r\n:"):
                return True
        return False



import time
from threading import Thread, Event
from time import sleep

import numpy as np
from karabo_bridge import Client as KbClient

from kamzik3 import units, DeviceError
from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.snippets.snippetDataAccess import get_from_dict, set_in_dict
from kamzik3.snippets.snippetsDecorators import expose_method


class DeviceKaraboBridge(Device):
    karabo_client = None

    def __init__(self, host, port, device_id=None, config=None):
        self.host = host
        self.port = port
        self.stop_auto_refresh = Event()
        self.counter_running = False
        self.stop_auto_refresh.clear()
        super(DeviceKaraboBridge, self).__init__(device_id, config)
        self.connect()

    def _init_attributes(self):
        Device._init_attributes(self)
        self.create_attribute(ATTR_REFRESHING, default_value=False, readonly=True, default_type=np.bool)

    def connect(self, *args):
        """
        Call only this function to connect devices to port / socket / library / ...
        :param args: connect attributes
        """
        self.connecting = True
        self.karabo_client = KbClient('tcp://{}:{}'.format(self.host, self.port))
        self.handle_configuration_event()
        self.connected = True
        self.connecting = False

    def handle_configuration(self):
        start_at = time.time()
        self._update_attributes()
        self._config_commands()
        self._config_attributes()
        self.start_polling()
        self.set_status(STATUS_CONFIGURED)
        self.auto_refresh()
        self.logger.info(u"Device configuration took {} sec.".format(time.time() - start_at))

    def _update_attributes(self):
        """
        To make things general, I'm just getting one next() call from Karabo bridge.
        Saving for now only floats and integer values.
        Ignoring numpy arrays.
        :return:
        """
        data, metadata = self.karabo_client.next()

        for key, value in data.items():
            self.attributes[key] = {}
            decimals = 0

            for sub_key, sub_value in value.items():
                if isinstance(sub_value, int):
                    default_type = np.int64
                elif isinstance(sub_value, float):
                    default_type = np.float64
                    decimals = 6
                else:
                    continue
                self.create_attribute(sub_key, key, default_value=sub_value, readonly=True, default_type=default_type,
                                      decimals=decimals)

        self.create_attribute(ATTR_MIN_UPDATE_TIMEOUT, default_value=0, default_type=np.float64, unit="ms", decimals=0)
        self.create_attribute(ATTR_TRAIN_TIMEOUT, default_value=100, default_type=np.float64, unit="ms", decimals=0)
        self.create_attribute(ATTR_REMOTE_TIMESTAMP, default_type=np.float64, unit="sec", readonly=True, decimals=6)
        self.create_attribute(ATTR_LOCAL_TIMESTAMP, default_type=np.float64, unit="sec", readonly=True, decimals=6)
        self.create_attribute(ATTR_DELTA_TIMESTAMP, default_value=0, default_type=np.int64, unit="ms", readonly=True,
                              decimals=6)
        self.create_attribute(ATTR_COUNTER, default_value=0, default_type=np.uint16, set_value_when_set_function=False,
                              set_function=self.count_trains)
        self.create_attribute(ATTR_LOCAL_TRAIN_ID, default_value=0, default_type=np.int64, readonly=True)
        self.create_attribute(ATTR_REMOTE_TRAIN_ID, default_value=0, default_type=np.int64, readonly=True)
        self.create_attribute(ATTR_TRAIN_OFFSET, default_value=0, default_type=np.int64)

        self.update_metadata(list(metadata.values())[0])

    def _set(self, attribute, value):
        """
        This sets Device attribute.
        Use this function when You want to set attribute by tuple or list key.
        Example: Device.set((ATTR_STATUS, VALUE), STATUS_IDLE)
        Attribute value is pushed into server if Device is connected on any.
        To reduce amount of pushed attributes we check if value is different from previous one.
        :param attribute: tuple, list, str
        :param value: mixed
        :return: None
        """
        token = 0
        if isinstance(attribute, (tuple, list)):
            current_value = get_from_dict(self.attributes, attribute)
            if isinstance(value, np.ndarray):
                if np.array_equal(value, current_value):
                    return
            elif value == current_value and attribute[0] != ATTR_COUNTER:
                return

            set_in_dict(self.attributes, attribute, value)
            if attribute[-1] == VALUE:
                # Get new value possibly affected by offset and factor
                value = get_from_dict(self.attributes, attribute)
            token = "{}.{}".format(TOKEN_ATTRIBUTE, ".".join(attribute))
        else:
            current_value = self.attributes[attribute]
            if isinstance(value, np.ndarray):
                if np.array_equal(value, current_value):
                    return
            elif value == current_value and attribute[0] != ATTR_COUNTER:
                return

            self.attributes[attribute] = value
            if attribute == VALUE:
                # Get new value possibly affected by offset and factor
                value = self.attributes[attribute]
            token = "{}.{}".format(TOKEN_ATTRIBUTE, attribute)

        if token and self.device_server is not None and not isinstance(value, np.ndarray):
            self.device_server.push_message(self.device_id, (attribute, value), token)

    def command(self, command, callback=None, with_token=False, returning=True):
        raise DeviceError("Device does not accept any commands.")

    def count_trains(self, counts):
        self.reset_counter()
        start_train = self.get_value(ATTR_LOCAL_TRAIN_ID)

        def counter(counts):
            self.set_status(STATUS_BUSY)
            counted = 0
            self.counter_running = True
            while counted < counts and self.counter_running:
                counted = self.get_value(ATTR_LOCAL_TRAIN_ID) - start_train
                self.set_raw_value(ATTR_COUNTER, counted)
                sleep(0.02)
            self.set_raw_value(ATTR_COUNTER, counts)
            self.counter_running = False
            self.set_status(STATUS_IDLE)

        Thread(target=counter, args=[counts]).start()

    @expose_method()
    def reset_counter(self):
        self.set_value(ATTR_COUNTER, 0)

    @expose_method()
    def refresh(self):
        """
        Get one next() call from Karabo bridge
        :return:
        """
        start_time = time.time()
        self.set_attribute([ATTR_REFRESHING, VALUE], True)

        def _thread():
            self.karabo_next()
            end_time = time.time() - start_time
            self.set_value(ATTR_LATENCY, end_time * 1e3)
            self.set_attribute([ATTR_REFRESHING, VALUE], False)

        Thread(target=_thread).start()

    @expose_method()
    def auto_refresh(self):
        """
        Start routine which will automatically call next() until stopped.
        To save CPU and potentially bandwidth set ATTR_MIN_UPDATE_TIMEOUT attribute.
        Take into account, that next() is blocking!
        :return:
        """
        self.reset_counter()
        if self.get_attribute([ATTR_REFRESHING, VALUE]):
            return

        self.logger.info("Starting auto refresh routine")
        self.set_attribute([ATTR_REFRESHING, VALUE], True)
        self.stop_auto_refresh.clear()

        def _thread():
            while not self.stop_auto_refresh.is_set():
                start_time = time.time()
                self.karabo_next()
                end_time = time.time() - start_time
                sleep_for = (self[ATTR_MIN_UPDATE_TIMEOUT][VALUE] - end_time * 1e3) / 1e3
                if sleep_for > 0:
                    sleep(sleep_for)
                self.set_value(ATTR_LATENCY, end_time * 1e3)

            self.set_attribute([ATTR_REFRESHING, VALUE], False)
            self.logger.info("Auto refresh routine stopped")

        Thread(target=_thread).start()

    def karabo_next(self):
        """
        Get one next() call from Karabo bridge.
        Take into account, that next() is blocking!
        :return:
        """
        data, metadata = self.karabo_client.next()
        for key, value in data.items():
            for sub_key, sub_value in value.items():
                if isinstance(sub_value, (int, float)):
                    self.set_value([key, sub_key], sub_value)
        self.update_metadata(list(metadata.values())[0])

    def update_metadata(self, metadata_values):
        """
        This routine is important to get and correct train ID from Karabo bridge.
        ATTR_REMOTE_TRAIN_ID - Karabo train ID
        ATTR_REMOTE_TIMESTAMP - Karabo timestamp
        ATTR_LOCAL_TIMESTAMP - Local computer timestamp
        ATTR_DELTA_TIMESTAMP - Delta timestamp (ATTR_LOCAL_TIMESTAMP - ATTR_REMOTE_TIMESTAMP)
        ATTR_TRAIN_OFFSET - User configurable train offset
        ATTR_LOCAL_TRAIN_ID - Interpolated train id using delta timestamp.
                              train_id_delta = ATTR_DELTA_TIMESTAMP / ATTR_TRAIN_TIMEOUT
                              (ATTR_REMOTE_TRAIN_ID + train_id_delta + ATTR_TRAIN_OFFSET)
        :param metadata_values: metadata from Karabo bridge
        :return:
        """
        remote_timestamp, local_timestamp = np.float64(metadata_values["timestamp"]), time.time()
        time_delta = (local_timestamp - remote_timestamp) * 1e3
        train_id_delta = units.Quantity(time_delta, "ms") / self[ATTR_TRAIN_TIMEOUT].value()
        self.set_attribute((ATTR_REMOTE_TRAIN_ID, VALUE), metadata_values["timestamp.tid"])
        self.set_attribute((ATTR_REMOTE_TIMESTAMP, VALUE), remote_timestamp)
        self.set_attribute((ATTR_LOCAL_TIMESTAMP, VALUE), np.float64(time.time()))
        self.set_attribute((ATTR_DELTA_TIMESTAMP, VALUE), time_delta)
        self.set_attribute((ATTR_LOCAL_TRAIN_ID, VALUE),
                           int(round(
                               metadata_values["timestamp.tid"] + train_id_delta.m + self[ATTR_TRAIN_OFFSET][VALUE])))

    @expose_method()
    def stop(self):
        """
        Stop auto refresh routine by setting stop_auto_refresh flag.
        :return:
        """
        self.logger.info("Stopping auto refresh routine")
        self.counter_running = False
        self.stop_auto_refresh.set()

    def close(self):
        self.stop()
        return super().close()



import collections
from threading import Lock

import pint
import pyqtgraph as pg
from PyQt5.QtCore import pyqtSignal, pyqtSlot

from kamzik3.snippets.snippetsYaml import YamlSerializable

from kamzik3 import units
from kamzik3.constants import *

pg.setConfigOption(u'background', u'w')
pg.setConfigOption(u'foreground', u"k")


class AttributePlot(pg.PlotWidget, YamlSerializable):

    sig_repaint = pyqtSignal()

    def __init__(self, attribute, buffer_size, plot=None, parent=None, background='#fff', **kwargs):
        super().__init__(parent, background, **kwargs)
        self.current_point = 0
        self.attribute = attribute
        self.x_buffer = collections.deque([], maxlen=buffer_size)
        self.y_buffer = collections.deque([], maxlen=buffer_size)
        try:
            self.base_unit = kwargs.get("unit", attribute[UNIT])
        except TypeError:
            self.base_unit = ""

        pen1 = pg.mkPen(color=(0, 0, 255))
        pen2 = pg.mkPen(color=(0, 0, 0))
        self.value_curve = pg.PlotDataItem(self.x_buffer, self.y_buffer, pen=pen1, antialias=False, symbol=u"s",
                                           symbolSize=2, symbolPen=pen2)
        if plot is None:
            self.setLabel(u"left", kwargs.get("left_label", ""), units=self.base_unit)
            self.setLabel(u"bottom", "point")
            self.showGrid(x=True, y=True, alpha=0.1)
            self.addItem(self.value_curve)
        else:
            plot.setLabel(u"left", kwargs.get("left_label", ""), units=self.base_unit)
            plot.setLabel(u"bottom", "point")
            plot.showGrid(x=True, y=True, alpha=0.1)
            plot.addItem(self.value_curve)
        self.sig_repaint.connect(self.slot_set_data)

    def yaml_mapping(self):
        mapping = super().yaml_mapping()
        del mapping["parent"]
        return mapping

    @pyqtSlot()
    def slot_set_data(self):
        self.value_curve.setData(self.x_buffer, self.y_buffer)

    def update(self, current_point=None, current_value=None):
        if current_point is None:
            self.current_point += 1
        else:
            self.current_point = current_point
        if current_value is None:
            try:
                current_value = self.attribute.value().to(self.base_unit).m
            except (AttributeError, pint.errors.DimensionalityError):
                # Value is None, ignore this point
                return
        else:
            assert isinstance(current_value, units.Quantity)
            current_value = current_value.to(self.base_unit).m
        try:
            self.y_buffer.append(current_value)
            self.x_buffer.append(self.current_point)
        except Exception as e:
            print(e, current_point, current_value)
            return
        self.sig_repaint.emit()

    def close(self):
        self.attribute = None
        self.current_point = 0
        self.reset()
        super().close()

    def reset(self):
        self.x_buffer.clear()
        self.y_buffer.clear()
        self.value_curve.setData([], [])



import logging
import statistics
import sys
import traceback
from logging.handlers import TimedRotatingFileHandler
from time import time, perf_counter_ns, sleep

import numpy

from kamzik3.snippets.snippetsTimer import PreciseCallbackTimer

base_log_formatter = logging.Formatter('%(asctime)s, %(name)s, %(levelname)s, %(message)s', datefmt='%Y-%m-%d %H:%M:%S')


def set_file_handler(logger, log_output_dir, formatter=base_log_formatter):
    handler = logging.FileHandler(log_output_dir)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return handler


def set_rotating_file_handler(logger, log_output_dir, formatter=base_log_formatter):
    handler = TimedRotatingFileHandler(log_output_dir, when="midnight")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return handler


def get_console_handler(log_level=logging.DEBUG, formatter=base_log_formatter):
    console_handler = logging.StreamHandler()
    console_handler.setLevel(log_level)
    console_handler.setFormatter(formatter)
    return console_handler


def set_sys_exception_handler(logger):
    def handle_exception(exc_type, exc_value, exc_traceback):
        logger.error("Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback))
        sys.__excepthook__(exc_type, exc_value, exc_traceback)

    sys.excepthook = handle_exception


def print_exception():
    print(traceback.format_exc())


class MeasureTiming:

    def __init__(self, title, timeout, save_file, measure_time=10000):
        self.spacing = []
        self.error = []
        self.measured_values = []
        self.title = title
        self.measure_time = int(measure_time)
        self.samples = 0
        self.running = False
        self.timeout = timeout
        self.timer = None
        self.save_file = save_file
        self.reference_time_perf = perf_counter_ns()

    def start_measurement(self):
        self.spacing = []
        self.error = []
        self.samples = 0
        self.running = True
        self.timer = PreciseCallbackTimer(self.measure_time + 10*self.timeout, self.plot)
        self.reference_time_perf = perf_counter_ns()
        self.timer.start()

    def measure(self, data=None):
        if not self.running:
            return
        t_diff = (perf_counter_ns() - self.reference_time_perf) * 1e-6
        self.measured_values.append(str((perf_counter_ns(), t_diff, data)))
        self.reference_time_perf = perf_counter_ns()
        self.spacing.append(t_diff)
        self.error.append(abs(t_diff - self.timeout))
        self.samples += 1

    def plot(self):
        self.running = False
        self.timer.stop()
        sleep(0.5)
        self.spacing = self.spacing[5:-5]
        self.error = self.error[5:-5]
        stat_data = []
        stat_data.append(self.title)
        stat_data.append(str((time(), "Frequency:", self.samples, "Median:", statistics.median(self.spacing),
              "Mean:", statistics.mean(self.spacing), "Min:", min(self.spacing), "Max:", max(self.spacing), "Std:",
              numpy.std(self.spacing))))
        stat_data.append(str(("Median error:", statistics.median(self.error), "Mean error:", statistics.mean(self.error), "Min:",
              min(self.error),
              "Max:",
              max(self.error),
              "Std:", numpy.std(self.error))))
        print(stat_data)
        with open("./collected_data.txt", "a+") as fp:
            fp.write("\r\n".join(stat_data))
            fp.write("\r\n".join(self.measured_values))

        import matplotlib.pyplot as plt
        plt.subplot(2, 1, 1)
        plt.scatter(range(len(self.spacing)), self.spacing, s=2, c="r")
        plt.title(self.title)
        plt.xlabel('samples [count]')
        plt.ylabel('spacing [ms]')

        plt.subplot(2, 1, 2)
        plt.scatter(range(len(self.error)), self.error, s=2, c="r")
        plt.xlabel('samples [count]')
        plt.ylabel('error [ms]')

        plt.savefig(self.save_file)



import json

from kamzik3 import units


class JsonKamzikEncoder(json.JSONEncoder):

    def default(self, o):
        if isinstance(o, units.Quantity):
            return {"pint-unit-quantity": o.to_tuple()}
        return o


def JsonPintHook(dct):
    if 'pint-unit-quantity' in dct:
        return units.Quantity.from_tuple(dct['pint-unit-quantity'])
    return dct


import sys
from time import perf_counter_ns

from kamzik3.snippets.snippetsInterfaces import CallbackTimerInterface


class CallbackTimer(CallbackTimerInterface):

    def run(self):
        if self.with_correction:
            time_reference = perf_counter_ns()
            while not self.stopped.wait(self.timeout - (perf_counter_ns() - time_reference) * 1e-9):
                self.callback()
                time_reference += self.timeout * 1e9
        else:
            while not self.stopped.wait(self.timeout):
                self.callback()


if sys.platform == "win32":
    # If more precise timer is requested for windows.
    # Problem of windows is that it's running on lower clock compared to Linux platform.
    # We can get only to resolution of 10ms.
    # We have to increase clock of CPU.
    # THIS CHANGES CLOCK GLOBALLY !!!!!!!!
    from kamzik3.snippets.snippetsWin32 import CallbackTimerWOC


    class PreciseCallbackTimer(CallbackTimerWOC):
        pass
else:
    # For Linux we use same timer since there is no problem with clock precision.
    # We can get up to 1us resolution.
    class PreciseCallbackTimer(CallbackTimer):
        pass
